set define '^' verify off

prompt ... Create APEX REST Source Catalog views

-----------------------------------------------------
-- W E B  S O U R C E  S E R V I C E  C A T A L O G S
--

--------------------------------------------------------------------------------

prompt ...apex_rest_catalog_groups

create or replace view apex_rest_catalog_groups
as
select 
    w.workspace,
    w.workspace_display_name,
    --
    g.id                                  catalog_group_id,
    g.name                                catalog_group_name,
    --
    (select count(c.id)
       from wwv_web_src_catalogs c
      where catalog_group_id = g.id )     catalogs_count,
    --
    g.last_updated_by,
    g.last_updated_on
from wwv_web_src_catalog_groups g,
     wwv_flow_current_workspaces w
where w.workspace_id       = g.security_group_id
/

comment on table apex_rest_catalog_groups is 'Groups of REST Source Catalogs';
comment on column apex_rest_catalog_groups.workspace                    is 'A work area mapped to one or more database schemas';
comment on column apex_rest_catalog_groups.workspace_display_name       is 'Display name for the workspace';
comment on column apex_rest_catalog_groups.catalog_group_id             is 'REST Source Catalog Group internal ID';
comment on column apex_rest_catalog_groups.catalog_group_name           is 'Name of the REST Source Catalog Group';
comment on column apex_rest_catalog_groups.catalogs_count               is 'Number of Catalogs within the group';
comment on column apex_rest_catalog_groups.last_updated_by              is 'APEX developer who made last update';
comment on column apex_rest_catalog_groups.last_updated_on              is 'Date of last update';

--------------------------------------------------------------------------------

prompt ...apex_rest_catalogs

create or replace view apex_rest_catalogs
as
select
    w.workspace,
    w.workspace_display_name,
    --
    g.id                                     catalog_group_id,
    g.name                                   catalog_group_name,
    --
    c.id                                     catalog_id,
    c.name                                   catalog_name,
    c.internal_name                          catalog_internal_name,
    c.description                            catalog_description,
    --
    c.endpoint_url                           url_endpoint, 
    c.token_server_url                       auth_url_endpoint,
    c.proxy_server,
    --
    (select count(cs.id)
       from wwv_web_src_catalog_services cs
      where catalog_id = c.id )              services_count,
    --
    cr.id                                    credential_id,
    cr.name                                  credential_name,
    cr.static_id                             credential_static_id,
    --
    c.format,
    --
    c.default_plugin_internal_name           default_plugin_code,
    decode( c.default_authentication_type,
            'BASIC',                     'Basic Authentication',
            'OAUTH2_CLIENT_CREDENTIALS', 'OAuth2 Client Credentials Flow',
            'OCI',                       'Oracle Cloud Infrastructure',
            'HTTP_HEADER',               'HTTP Header',
            'HTTP_QUERY_STRING',         'HTTP Query String'
    )                                        default_authentication_type,
    c.version,
    c.last_imported_at,
    c.last_import_file,
    --
    c.last_updated_by,
    c.last_updated_on
from wwv_web_src_catalogs c,
     wwv_web_src_catalog_groups g,
     wwv_flow_current_workspaces w,
     wwv_credentials cr
where w.workspace_id = c.security_group_id
  and g.id           = c.catalog_group_id
  and cr.id (+)      = c.credential_id 
/

comment on table apex_rest_catalogs is 'REST Source Catalogs';
comment on column apex_rest_catalogs.workspace                    is 'A work area mapped to one or more database schemas';
comment on column apex_rest_catalogs.workspace_display_name       is 'Display name for the workspace';
comment on column apex_rest_catalogs.catalog_group_id             is 'Catalog Group internal ID';
comment on column apex_rest_catalogs.catalog_group_name           is 'Name of the Catalog Group';
comment on column apex_rest_catalogs.catalog_id                   is 'REST Source Catalog internal ID';
comment on column apex_rest_catalogs.catalog_name                 is 'Name of the REST Source Catalog';
comment on column apex_rest_catalogs.catalog_internal_name        is 'Internal Name of the catalog';
comment on column apex_rest_catalogs.catalog_description          is 'Description of the catalog';
comment on column apex_rest_catalogs.services_count               is 'Number of services within the REST Source Catalog';
comment on column apex_rest_catalogs.url_endpoint                 is 'URL endpoint from which this catalog can be refreshed';
comment on column apex_rest_catalogs.auth_url_endpoint            is 'URL endpoint to use for authentication when the catalog is refreshed';
comment on column apex_rest_catalogs.proxy_server                 is 'Proxy Server to use when refreshing the catalog';
comment on column apex_rest_catalogs.credential_id                is 'Internal ID of the Web Credential to use for authentication when refreshing the catalog';
comment on column apex_rest_catalogs.credential_name              is 'Name of the Web Credential to use for authentication when refreshing the catalog';
comment on column apex_rest_catalogs.credential_static_id         is 'Static ID of the Web Credential to use for authentication when refreshing the catalog';
comment on column apex_rest_catalogs.format                       is 'The file format to use when refreshing the catalog';
--
comment on column apex_rest_catalogs.default_plugin_code          is 'Internal identifier for the plug-in to use by default for services within the catalog';
comment on column apex_rest_catalogs.default_authentication_type  is 'Type of authentication to use by default for services within the catalog';
comment on column apex_rest_catalogs.version                      is 'Version identifier for this catalog';
comment on column apex_rest_catalogs.last_imported_at             is 'Date when this catalog was last imported or refreshed';
comment on column apex_rest_catalogs.last_import_file             is 'The contents of the file which was last used to refresh this catalog';
comment on column apex_rest_catalogs.last_updated_by              is 'APEX developer who made last update';
comment on column apex_rest_catalogs.last_updated_on              is 'Date of last update';

--------------------------------------------------------------------------------

prompt ...apex_rest_catalog_plugins

create or replace view apex_rest_catalog_plugins
as
select 
    w.workspace,
    w.workspace_display_name,
    --
    c.id                                                 catalog_id,
    c.name                                               catalog_name,
    --
    (select count(cs.id)
       from wwv_web_src_catalog_services cs
      where catalog_id = c.id 
        and plugin_internal_name = p.plugin_internal_name)
                                                         usage_count,
    --
    p.id                                                 plugin_id,
    p.plugin_internal_name                               plugin_name,
    p.plugin_display_name                                plugin_display_name,
    p.plugin_version                                     plugin_version,
    p.plugin_export                                      export_file
from wwv_web_src_catalog_plugins p,
     wwv_web_src_catalogs        c,
     wwv_flow_current_workspaces w
where w.workspace_id = p.security_group_id
  and c.id           = p.catalog_id
/

comment on table apex_rest_catalog_plugins is 'REST Source Plug-Ins used by services in a REST Source Catalog';
comment on column apex_rest_catalog_plugins.workspace                    is 'A work area mapped to one or more database schemas';
comment on column apex_rest_catalog_plugins.workspace_display_name       is 'Display name for the workspace'; 
comment on column apex_rest_catalog_plugins.catalog_id                   is 'Internal ID of the REST Source Catalog which this plug-in belongs to';
comment on column apex_rest_catalog_plugins.catalog_name                 is 'Name of the REST Source Catalog which this plug-in belongs to';
comment on column apex_rest_catalog_plugins.usage_count                  is 'Number of services in the REST Source Catalog, which use this plug-in';
comment on column apex_rest_catalog_plugins.plugin_id                    is 'Internal identifier for this plug-in within a catalog';
comment on column apex_rest_catalog_plugins.plugin_name                  is 'Internal name for the plug-in, which is used to reference in REST Data Sources';
comment on column apex_rest_catalog_plugins.plugin_display_name          is 'Display Name of the plug-in';
comment on column apex_rest_catalog_plugins.plugin_version               is 'Version of the plug-in when it was exported';
comment on column apex_rest_catalog_plugins.export_file                  is 'Export file for of the plug-in, in APEX export (SQL) format';

--------------------------------------------------------------------------------

prompt ...apex_rest_catalog_services

create or replace view apex_rest_catalog_services
as
select 
    w.workspace,
    w.workspace_display_name,
    --
    c.id                                                 catalog_id,
    c.name                                               catalog_name,
    --
    s.id                                                 service_id,
    s.name                                               service_name,
    s.description                                        service_description,
    --
    s.base_url || s.service_path                         url_endpoint,
    s.auth_server_base_url || s.auth_server_url_path     auth_url_endpoint,
    --
    s.plugin_internal_name                               plugin_name,
    decode( s.authentication_type,
            'BASIC',                     'Basic Authentication',
            'OAUTH2_CLIENT_CREDENTIALS', 'OAuth2 Client Credentials Flow',
            'OCI',                       'Oracle Cloud Infrastructure',
            'HTTP_HEADER',               'HTTP Header',
            'HTTP_QUERY_STRING',         'HTTP Query String'
    )                                                    authentication_type,
    --
    s.details_json,
    --
    s.version,
    s.last_imported_at
from wwv_web_src_catalog_services s,
     wwv_web_src_catalogs c,
     wwv_flow_current_workspaces w
where w.workspace_id = s.security_group_id
  and c.id           = s.catalog_id
/

comment on table apex_rest_catalog_services is 'Services within a REST Source Catalog';
comment on column apex_rest_catalog_services.workspace                    is 'A work area mapped to one or more database schemas';
comment on column apex_rest_catalog_services.workspace_display_name       is 'Display name for the workspace'; 
comment on column apex_rest_catalog_services.catalog_id                   is 'Internal ID for the REST Source Catalog containing this service';
comment on column apex_rest_catalog_services.catalog_name                 is 'Name of the REST Source Catalog containing this service';
comment on column apex_rest_catalog_services.service_id                   is 'Internal ID for this service';
comment on column apex_rest_catalog_services.service_name                 is 'Catalog Service name';
comment on column apex_rest_catalog_services.service_description          is 'Description of the catalog service';
comment on column apex_rest_catalog_services.url_endpoint                 is 'URL endpoint for this catalog service';
comment on column apex_rest_catalog_services.auth_url_endpoint            is 'Token server endpoint, if the service uses OAuth authentication';
--
comment on column apex_rest_catalog_services.plugin_name                  is 'Internal name for the REST Source plug-in to use for ths service';
comment on column apex_rest_catalog_services.authentication_type          is 'Type of authentication to use for the service';
comment on column apex_rest_catalog_services.details_json                 is 'Service details, in JSON format';
comment on column apex_rest_catalog_services.version                      is 'Version identifier for this service';
comment on column apex_rest_catalog_services.last_imported_at             is 'Date when this service was last imported';

--------------------------------------------------------------------------------

prompt ...done creating APEX REST Source Catalog views

set define '^'
