set define '^' verify off
prompt ...wwv_flow_application_install
create or replace package wwv_flow_application_install as
--------------------------------------------------------------------------------
--
-- Copyright (c) Oracle Corporation 2010 - 2021. All Rights Reserved.
--
-- The APEX_APPLICATION_INSTALL package provides many methods to modify application
-- attributes during the Application Express application installation process.
--
-- Oracle Application Express provides two ways to import an application into an 
-- Application Express instance:
-- 1. Upload and installation of an application export file by using the web 
--    interface of Application Express.
-- 2. Execution of the application export file as a SQL script, typically in 
--    the command-line utility SQL*Plus.
-- 
-- Using the file upload capability of the web interface of Application Express, 
-- developers can import an application with a different application ID, different
-- workspace ID and different parsing schema. But when importing an application
-- by using a command-line tool like SQL*Plus, none of these attributes 
-- (application ID, workspace ID, parsing schema) can be changed without directly
-- modifying the application export file.
--
-- To view the install log, enter the following from the command-line tool, so 
-- the server outputs are displayed:
--
--   set serveroutput on unlimited
--
-- As more and more Application Express customers create applications which are
-- meant to be deployed by using command-line utilities or by using a non-web-based 
-- installer, they are faced with this challenge of how to import their application 
-- into an arbitrary workspace on any Application Express instance.
--
-- Another common scenario is in a training class when installing an application 
-- into 50 different workspaces that all use the same application export file. 
-- Today, customers  work around this by adding their own global variables to an 
-- application export file and then varying the values of these globals at installation 
-- time. However, this manual modification of the application export file (usually 
-- done with a post-export sed or awk script) should not be necessary.
--
-- Application Express 4.0 and higher includes the APEX_APPLICATION_INSTALL API.
-- This PL/SQL API provides many methods to set application attributes during 
-- the Application Express application installation process. All export files
-- in Application Express 4.0 and higher contain references to the values set
-- by the APEX_APPLICATION_INSTALL API. However, the methods in this API is only
-- used to override the default application installation behavior.
--
-- The table below lists the attributes that can be set by functions in this API.
--
-- Workspace ID:   Workspace ID of the imported application. See GET_WORKSPACE_ID
--                 and SET_WORKSPACE_ID.
-- Application ID: Application ID of the imported application.
--                 See GENERATE_APPLICATION_ID, GET_APPLICATION_ID, SET_APPLICATION_ID.
-- Offset:         Offset value used during application import. See GENERATE_OFFSET,
--                 GET_OFFSET, SET_OFFSET.
-- Schema:         The parsing schema (`owner`) of the imported application. See
--                 GET_SCHEMA, SET_SCHEMA.
-- Name:           Application name of the imported application.
--                 See GET_APPLICATION_NAME and SET_APPLICATION_NAME.
-- Alias:          Application alias of the imported application.
--                 See GET_APPLICATION_ALIAS and SET_APPLICATION_ALIAS.
-- Image Prefix:   The image prefix of the imported application.
--                 See GET_IMAGE_PREFIX and SET_IMAGE_PREFIX.
-- Proxy:          The proxy server attributes of the imported application.
--                 See GET_PROXY and SET_PROXY.
--
-- # Import Script Examples
--
-- Using the workspace `FRED_DEV` on the development instance, you generate an
-- application export of application 645 and save it as file `f645.sql`.
-- All examples in this section assume you are connected to SQL*Plus.
--
-- ## Import Application without Modification
--
-- To import this application back into the `FRED_DEV` workspace on the same
-- development instance using the same application ID:
--
--   @f645.sql
--
-- ## Import Application with Specified Application ID
--
-- To import this application back into the `FRED_DEV` workspace on the same
-- development instance, but using application ID 702:
--
--   begin
--       apex_application_install.set_application_id(702);
--       apex_application_install.generate_offset;
--       apex_application_install.set_application_alias( 'F' || apex_application_install.get_application_id );
--   end;
--   /
-- 
--   @645.sql
--
-- ## Import Application with Generated Application ID
--
-- To import this application back into the FRED_DEV workspace on the same 
-- development instance, but using an available application ID generated by 
-- Application Express:
--
--   begin
--       apex_application_install.generate_application_id;
--       apex_application_install.generate_offset;
--       apex_application_install.set_application_alias( 'F' || apex_application_install.get_application_id );
--   end;
--   /
-- 
--   @f645.sql
--
-- ## Import Application into Different Workspace using Different Schema
--
-- To import this application into the `FRED_PROD` workspace on the production instance, 
-- using schema `FREDDY`, and the workspace ID of `FRED_DEV` and `FRED_PROD` are different:
--
--   begin
--       apex_application_install.set_workspace('FRED_PROD');
--       apex_application_install.generate_offset;
--       apex_application_install.set_schema( 'FREDDY' );
--       apex_application_install.set_application_alias( 'FREDPROD_APP' );
--   end;
--   /
--
--   @f645.sql
--
-- ## Import into Training Instance for Three Different Workspaces
--
-- To import this application into the Training instance for 3 different workspaces:
--
--   begin
--       apex_application_install.set_workspace('TRAINING1');
--       apex_application_install.generate_application_id;
--       apex_application_install.generate_offset;
--       apex_application_install.set_schema( 'STUDENT1' );
--       apex_application_install.set_application_alias( 'F' || apex_application_install.get_application_id );
--   end;
--   /
-- 
--   @f645.sql
-- 
--   begin
--       apex_application_install.set_workspace('TRAINING2');
--       apex_application_install.generate_application_id;
--       apex_application_install.generate_offset;
--       apex_application_install.set_schema( 'STUDENT2' );
--       apex_application_install.set_application_alias( 'F' || apex_application_install.get_application_id );
--   end;
--   /
-- 
--   @f645.sql
-- 
--   begin
--       apex_application_install.set_workspace('TRAINING3');
--       apex_application_install.generate_application_id;
--       apex_application_install.generate_offset;
--       apex_application_install.set_schema( 'STUDENT3' );
--       apex_application_install.set_application_alias( 'F' || apex_application_install.get
--                                                     _application_id );
--   end;
--   /
--
--   @f645.sql
--------------------------------------------------------------------------------

--
--##############################################################################
-- Import Data Types
--
-- The section describes import data types used by the APEX_APPLICATION_INSTALL package.
--##############################################################################
--
--------------------------------------------------------------------------------
-- Defines the kind of install files.
--------------------------------------------------------------------------------
subtype t_file_type is pls_integer range 1 .. 5;
c_file_type_workspace           constant t_file_type := 1;
c_file_type_app                 constant t_file_type := 2;
c_file_type_websheet            constant t_file_type := 3; -- obsolete
c_file_type_plugin              constant t_file_type := 4;
c_file_type_css                 constant t_file_type := 5;

--------------------------------------------------------------------------------
-- Define the kind of application usage.
--------------------------------------------------------------------------------
subtype t_app_usage is pls_integer range 1..3;
c_app_usage_not_used            constant t_app_usage := 1;
c_app_usage_current_workspace   constant t_app_usage := 2;
c_app_usage_other_workspace     constant t_app_usage := 3;

--------------------------------------------------------------------------------
-- Specifies information in a source file that can be used to configure the installation.
--
-- Attributes:
-- * file_type:
-- * workspace_id:
-- * version:
-- * app_id:
-- * app_name:
-- * app_alias:
-- * app_owner:
-- * build_status:
-- * has_install_script:
-- * app_id_usage:
-- * app_alias_usage:
--------------------------------------------------------------------------------
type t_file_info is record (
     file_type          t_file_type,
     workspace_id       number,
     version            varchar2(10),
     app_id             number,
     app_name           varchar2(4000),
     app_alias          varchar2(4000),
     app_owner          varchar2(4000),
     build_status       varchar2(4000),
     has_install_script boolean,
     app_id_usage       t_app_usage,
     app_alias_usage    t_app_usage );

--==============================================================================
-- Use this function to set the workspace ID for the application to be imported.
--
-- Parameters
-- * p_workspace_id: The workspace ID.
--
-- Example:
-- 
-- For examples of this procedure call, see "Import Application into Different 
-- Workspace using Different Schema" and "Import into Training Instance for Three Different Workspaces".
--
-- See Also:
-- - SET_WORKSPACE
-- - GET_WORKSPACE_ID
--==============================================================================
procedure set_workspace_id(
    p_workspace_id in number );
--
--==============================================================================
-- This function is used to set the workspace ID for the application to be imported.
--
-- Parameters:
-- * p_workspace: The workspace name.
--
-- Example:
--
-- This example shows how to set workspace ID for workspace `FRED_PROD`.
--
--   begin
--       apex_application_install.set_workspace (
--           p_workspace => 'FRED_PROD' );
--   end;
--
-- See Also:
-- - GET_WORKSPACE_ID
-- - SET_WORKSPACE_ID
--==============================================================================
procedure set_workspace (
    p_workspace in varchar2 );
--
--==============================================================================
-- Use this function to get the workspace ID for the application to be imported.
--
-- Example:
-- 
-- The following example returns the value of the workspace ID value in the
-- APEX_APPLICATION_INSTALL package.
--
--   declare
--       l_workspace_id number;
--   begin
--       l_workspace_id := apex_application_install.get_workspace_id;
--   end;
--
-- See Also:
-- - SET_WORKSPACE_ID
-- - SET_WORKSPACE
--==============================================================================
function get_workspace_id
    return number;
--
--==============================================================================
-- Use this procedure to set the application ID of the application to be imported.
-- The application ID should either not exist in the instance or, if it does exist,
-- must be in the workspace where the application is being imported to. This number
-- must be a positive integer and must not be from the reserved range of
-- Application Express application IDs.
--
-- Parameters:
-- * p_application_id: This is the application ID. The application ID must be 
--                     a positive integer, and cannot be in the reserved range of
--                     application IDs (3000 - 8999). It must be less than 3000
--                     or greater than or equal to 9000.
--
-- Example:
--
-- For an example of this procedure call, see "Import Application with Specified Application ID."
--
-- See Also:
-- - GENERATE_APPLICATION_ID
-- - GET_APPLICATION_ID
--==============================================================================
procedure set_application_id(
    p_application_id in number );
--
--==============================================================================
-- Use this function to get the application ID of the application to be imported.
-- The application ID should either not exist in the instance or, if it does exist,
-- must be in the workspace where the application is being imported to.
--
-- Example:
--
-- The following example returns the value of the application ID value in the
-- APEX_APPLICATION_INSTALL package.
--
--   declare
--       l_id number;
--   begin
--       l_id := apex_application_install.get_application_id;
--   end;
--
-- See Also:
-- - SET_APPLICATION_ID
-- - GENERATE_APPLICATION_ID
--==============================================================================
function get_application_id
    return number;
--
--==============================================================================
-- This procedure generates an available application ID on the instance and sets
-- the application ID in APEX_APPLICATION_INSTALL.
--
-- Example:
-- 
-- For an example of this procedure call, see "Import Application with Generated Application ID"
-- and "Import into Training Instance for Three Different Workspaces".
--
-- See Also:
-- - SET_APPLICATION_ID
-- - GET_APPLICATION_ID
--==============================================================================
procedure generate_application_id;
--
--==============================================================================
-- This procedure sets the offset value used during application import. Use the
-- offset value to ensure that the metadata for the Application Express application 
-- definition does not collide with other metadata on the instance. For a new 
-- application installation, it is usually sufficient to call the GENERATE_OFFSET
-- procedure to have Application Express generate this offset value for you.
--
-- Parameters:
-- * p_offset: The offset value. The offset must be a positive integer. In most
--             cases you do not need to specify the offset, and instead, call
--             GENERATE_OFFSET, which generates a large random value and then
--             set it in the APEX_APPLICATION_INSTALL package.
--
-- Example:
--
-- The following example generates a random number from the database and uses
-- this as the offset value in APEX_APPLICATION_INSTALL.
--
--   declare
--       l_offset number;
--   begin
--       l_offset := dbms_random.value(100000000000, 999999999999);
--       apex_application_install.set_offset( p_offset => l_offset );
--   end;
--
-- See Also:
-- - GET_OFFSET
-- - GENERATE_OFFSET
--==============================================================================
procedure set_offset(
    p_offset in number );
--
--==============================================================================
-- This procedure generates the offset value used during application import.
-- Use the offset value to ensure that the metadata for the Application Express
-- application  definition does not collide with other metadata on the instance.
-- For a new application installation, it is usually sufficient to call this
-- procedure to have Application Express generate this offset value for you.
--
-- Example: 
-- 
-- For examples of this procedure call, see "Import Application with Specified Application ID",
-- "Import Application with Generated Application ID" and
-- "Import into Training Instance for Three Different Workspaces".
--
-- See Also:
-- - GET_OFFSET
-- - SET_OFFSET
--==============================================================================
procedure generate_offset;
--
--==============================================================================
-- Use function to get the offset value used during the import of an application.
--
-- Example:
--
-- The following example returns the value of the application offset value in the 
-- APEX_APPLICATION_INSTALL package.
--
--   declare
--       l_offset number;
--   begin
--       l_offset := apex_application_install.get_offset;
--   end;
--
-- See Also:
-- - SET_OFFSET
-- - GENERATE_OFFSET
--==============================================================================
function get_offset
    return number;
--
--==============================================================================
-- Use this procedure to set the parsing schema (`owner`) of the Application Express
-- application. The database user of this schema must already exist, and this
-- schema name must already be mapped to the workspace used to import the application.
--
-- Example:
--
-- For examples of this procedure call, see "Import Application into Different Workspace using Different Schema"
-- and "Import into Training Instance for Three Different Workspaces".
--
-- See Also:
-- GET_SCHEMA
--==============================================================================
procedure set_schema(
    p_schema in varchar2 );
--
--==============================================================================
-- Use this function to get the parsing schema (`owner`) of the Application
-- Express application.
--
-- Example:
--
-- The following example returns the value of the application schema in the
-- APEX_APPLICATION_INSTALL package.
--
--   declare
--       l_schema varchar2(30);
--   begin
--       l_schema := apex_application_install.get_schema;
--   end;
--
-- See Also:
-- SET_SCHEMA
--==============================================================================
function get_schema
    return varchar2;
--
--==============================================================================
-- This procedure sets the application name of the import application.
--
-- Parameters:
-- * p_application_name: This is the application name. The application name cannot
--                       be null and cannot be longer than 255 characters.
--
-- Example:
--
-- The following example sets the application name in APEX_APPLICATION_INSTALL
-- to `Executive Dashboard`.
-- 
--   declare
--       c_name constant varchar2(255) := 'Executive Dashboard';
--   begin
--       apex_application_install.set_application_name(
--           p_application_name => c_name );
--   end;
--
-- See Also:
-- GET_APPLICATION_NAME
--==============================================================================
procedure set_application_name(
    p_application_name in varchar2 );
--
--==============================================================================
-- This function gets the application name of the import application.
--
-- Example:
--
-- The following example returns the value of the application name value in the
-- APEX_APPLICATION_INSTALL package.
--
--   declare
--       l_application_name varchar2(255);
--   begin
--       l_application_name := apex_application_install.get_application_name;
--   end;
--
-- See Also:
-- SET_APPLICATION_NAME
--==============================================================================
function get_application_name
    return varchar2;
--
--==============================================================================
-- This procedure sets the application alias for the application to be imported.
-- This is only used if the application to be imported has an alias specified.
-- An application alias must be unique within a workspace and it is recommended
-- to be unique within an instance.
--
-- Parameters:
-- * p_application_alias: The application alias. The application alias is an
--                        alphanumeric identifier. It cannot exceed 255 characters,
--                        must unique within a workspace and, ideally, is unique
--                        within an entire instance.
--
-- Example:
--
-- For examples of this procedure call, see "Import Application with Specified Application ID",
-- "Import Application with Generated Application ID",
-- "Import Application into Different Workspace using Different Schema" and
-- "Import into Training Instance for Three Different Workspaces."
--
-- See Also:
-- GET_APPLICATION_ALIAS
--==============================================================================
procedure set_application_alias(
    p_application_alias in varchar2 );
--
--==============================================================================
-- This function gets the application alias for the application to be imported. 
-- This is only used if the application to be imported has an alias specified.
-- An application alias must be unique within a workspace and it is recommended
-- to be unique within an instance.
--
-- Example:
--
-- The following example returns the value of the application alias value in the
-- APEX_APPLICATION_INSTALL package. The application alias cannot be more than
-- 255 characters.
--
--   declare
--       l_alias varchar2(255);
--   begin
--       l_alias := apex_application_install.get_application_alias;
--   end;
--
-- See Also:
-- SET_APPLICATION_ALIAS
--==============================================================================
function get_application_alias
    return varchar2;
--
--==============================================================================
-- This procedure sets the image prefix of the import application. Most Application
-- Express instances use the default image prefix of `/i/`.
--
-- Parameters:
-- * p_image_prefix: The image prefix. It can be a fully qualified domain, like
--                   a CDN or another web server, or just a path.
--
-- Example:
--
-- The following example sets the value of the image prefix variable in 
-- APEX_APPLICATION_INSTALL.
--
--   declare
--       c_prefix constant varchar2(255) := '/i/';
--   begin
--       apex_application_install.set_image_prefix( p_image_prefix => c_prefix );
--   end;
--==============================================================================
procedure set_image_prefix(
    p_image_prefix in varchar2 );
--
--==============================================================================
-- This function gets the image prefix of the import application. Most Application
-- Express instances use the default image prefix of `/i/`.
--
-- Example:
--
-- The following example returns the value of the application image prefix in the
-- APEX_APPLICATION_INSTALL package. The application image prefix cannot be more
-- than 255 characters.
--
--   declare
--       l_image_prefix varchar2(255);
--   begin
--       l_image_prefix := apex_application_install.get_image_prefix;
--   end;
--
-- See Also:
-- SET_IMAGE_PREFIX
--==============================================================================
function get_image_prefix
    return varchar2;
--
--==============================================================================
-- Use this procedure to set the proxy server attributes of an application to be imported.
--
-- Parameters:
-- * p_proxy:            The proxy server. There is no default value. The proxy
--                       server cannot be more than 255 characters and should
--                       not include any protocol prefix such as
--                       http://. A sample value might be: `www-proxy.example.com`
-- * p_no_proxy_domains: Comma delimited list of domains for which the proxy server should 
--                       not be used. There is no default value.
--
-- Example:
-- 
-- The following example sets the value of the proxy variable in APEX_APPLICATION_INSTALL.
--
--   declare
--       c_proxy constant varchar2(255) := 'www-proxy.example.com'
--   begin
--       apex_application_install.set_proxy( p_proxy => c_proxy );
--   end;
--
-- See Also:
-- - GET_PROXY
-- - GET_NO_PROXY_DOMAINS
--==============================================================================
procedure set_proxy(
    p_proxy            in varchar2,
    p_no_proxy_domains in varchar2 default null );
--
--==============================================================================
-- Use this function to get the proxy server attribute of an application to be imported.
--
-- Example:
--
-- The following example returns the value of the proxy server attribute in the
-- APEX_APPLICATION_INSTALL package. The proxy server attribute cannot be more 
-- than 255 characters.
--
--   declare
--       l_proxy varchar2(255);
--   begin
--       l_proxy := apex_application_install.get_proxy;
--   end;
--
-- See Also:
-- SET_PROXY
--==============================================================================
function get_proxy
    return varchar2;
--
--==============================================================================
-- Use this function to get the no proxy domains attribute of an application to be imported.
--
-- Returns:
-- Comma delimited list of domains for which the proxy server should not be used.
--
-- See Also:
-- SET_PROXY
--==============================================================================
function get_no_proxy_domains
    return varchar2;
--
--==============================================================================
-- Use this procedure to set the Base URL and the HTTPS Host attributes for remote
-- servers of the imported application. Remote Servers are identified by their Static ID.
--
-- Parameters:
-- * p_static_id:  Static ID to reference the remote server object.
-- * p_base_url:   New Base URL to use for this remote server object.
-- * p_https_host: New HTTPS Host Property to use for this remote server object.
--                 Only relevant when the base URL is `https://` and the database 
--                 version is 12.2 or greater.
--
-- Example:
--
--   begin
--       apex_application_install.set_remote_server(
--           p_static_id => 'MY_REMOTE_SERVER',
--           p_base_url => 'http://production.example.com' );
--   end;
--
-- See Also:
-- - GET_REMOTE_SERVER_BASE_URL
-- - GET_REMOTE_SERVER_HTTPS_HOST
--==============================================================================
procedure set_remote_server(
    p_static_id     in varchar2,
    p_base_url      in varchar2,
    p_https_host    in varchar2 default null,
    p_ords_timezone in varchar2 default null );
--
--==============================================================================
-- Use this function to get the Base URL property to be used for a given remote 
-- server during application import.
--
-- Parameters:
-- * p_static_id: Static ID to reference the remote server object.
--
-- Example:
--
--   declare
--       l_base_url varchar2(255);
--   begin
--       l_base_url := apex_application_install.get_remote_server_base_url( 'MY_REMOTE_SERVER' );
--   end;
--
-- See Also:
-- SET_REMOTE_SERVER
--==============================================================================
function get_remote_server_base_url(
    p_static_id in varchar2 )
    return varchar2;
--
--==============================================================================
-- Use this function to get the HTTPS Host property to be used for a given remote 
-- server during application import.
--
-- Parameters:
-- * p_static_id: Static ID to reference the remote server object.
--
-- Example:
--
--   declare
--       l_https_host varchar2(255);
--   begin
--       l_https_host := apex_application_install.get_remote_server_https_host( 'MY_REMOTE_SERVER' );
--   end;
--
-- See Also:
-- SET_REMOTE_SERVER
--==============================================================================
function get_remote_server_https_host(
    p_static_id in varchar2 )
    return varchar2;
--
--==============================================================================
-- Internal:
--==============================================================================
function get_remote_server_ords_tz(
    p_static_id in varchar2 )
    return varchar2;
--
--==============================================================================
-- This procedure sets the automatic install of supporting objects value used 
-- during application import. This setting is valid only for command line installs.
-- If the value is set to TRUE and the application export contains supporting 
-- objects, it automatically installs or upgrades the supporting objects when
-- an application imports from the command line.
--
-- Parameters:
-- * p_auto_install_sup_obj: The automatic install of supporting objects BOOLEAN value.
--
-- Example:
--
-- The following example gets the automatic install of supporting objects setting.
-- If it is not set to install automatically, it sets to TRUE to override export
-- file settings of automatic install of supporting objects.
--
--   begin
--       apex_application_install.set_auto_install_sup_obj(
--           p_auto_install_sup_obj => true );
--   end;
--
-- See Also:
-- GET_AUTO_INSTALL_SUP_OBJ
--==============================================================================
procedure set_auto_install_sup_obj(
    p_auto_install_sup_obj in boolean );
--
--==============================================================================
-- Use this function to get the automatic install of supporting objects setting
-- used during the import of an application. This setting is valid only for
-- command line installs. If the setting is set to TRUE and the application export 
-- contains supporting objects, it automatically installs or upgrades the supporting
-- objects when an application imports from the command line.
--
-- Example:
--
-- The following example returns the value of automatic install of supporting
-- objects setting in the APEX_APPLICATION_INSTALL package.
--
--   declare
--       l_auto_install_sup_obj boolean;
--   begin
--       l_auto_install_sup_obj := apex_application_install.get_auto_install_sup_obj;
--   end;
--
-- See Also:
-- SET_AUTO_INSTALL_SUP_OBJ
--==============================================================================
function get_auto_install_sup_obj
    return boolean;
--
--==============================================================================
-- Internal:
-- Static Application File Prefix
--
-- See Also:
-- GET_STATIC_APP_FILE_PREFIX
--==============================================================================
procedure set_static_app_file_prefix(
    p_file_prefix in varchar2 );
--
--==============================================================================
-- Internal:
--
-- See Also:
-- SET_STATIC_APP_FILE_PREFIX
--==============================================================================
function get_static_app_file_prefix
    return varchar2;
--
--==============================================================================
-- Internal:
-- Static Theme File Prefix
--
-- See Also:
-- GET_STATIC_THEME_FILE_PREFIX
--==============================================================================
procedure set_static_theme_file_prefix(
    p_theme_number in number,
    p_file_prefix  in varchar2 );
--
--==============================================================================
-- Internal:
--
-- See Also:
-- SET_STATIC_THEME_FILE_PREFIX
--==============================================================================
function get_static_theme_file_prefix(
    p_theme_number in number )
    return varchar2;
--
--==============================================================================
-- Internal:
-- Static Plugin File Prefix
--
-- See Also:
-- GET_STATIC_PLUGIN_FILE_PREFIX
--==============================================================================
procedure set_static_plugin_file_prefix(
    p_plugin_type in varchar2,
    p_plugin_name in varchar2,
    p_file_prefix in varchar2 );
--
--==============================================================================
-- Internal:
--
-- See Also:
-- SET_STATIC_PLUGIN_FILE_PREFIX
--==============================================================================
function get_static_plugin_file_prefix(
    p_plugin_type in varchar2,
    p_plugin_name in varchar2 )
    return varchar2;
--
--==============================================================================
-- This procedure preserves sessions associated with the application on upgrades.
--
-- Parameters
-- * p_keep_sessions: FALSE is the default value. TRUE if sessions should be preserved,
--                    FALSE if they should be deleted.
--                    The instance parameter `KEEP_SESSIONS_ON_UPGRADE` controls
--                    the default behavior. If it is `N` (the default), sessions
--                    will be deleted.
--
-- Example
-- 
-- The following example installs application 100 in workspace `FRED_PROD` and keep 
-- session state.
--
--   begin
--       apex_application_install.set_workspace(p_workspace => 'FRED_PROD');
--       apex_application_install.set_keep_sessions(p_keep_sessions => true);
--   end;
--   /
--   @f100.sql
--
-- See Also:
-- - GET_KEEP_SESSIONS
-- - Instance parameter `KEEP_SESSIONS_ON_UPGRADE`.
--==============================================================================
procedure set_keep_sessions (
    p_keep_sessions in boolean );
--
--==============================================================================
-- This function finds out if sessions and session state will be preserved or 
-- deleted on upgrades.
--
-- Example:
-- 
-- The following example shows whether print sessions will be kept or deleted.
--
--   begin
--       dbms_output.put_line (
--           case when apex_application_install.get_keep_sessions then 'sessions will be kept'
--           else 'sessions will be deleted'
--           end );
--   end;
--
-- See Also:
-- SET_KEEP_SESSIONS
--==============================================================================
function get_keep_sessions
    return boolean;
--
--==============================================================================
-- Use this procedure to override the active authentication scheme for the 
-- applications that are about to be installed.
--
-- Parameters
-- * p_name: The name of the authentication scheme to be activated This new
--           authentication scheme must exist in the application. If null,
--           the active authentication scheme will remain unchanged.
--
-- Example
--
-- Activate authentication scheme `SSO-Production` and install application
-- `f100.sql`, then reset the override for `f101.sql` to keep its active scheme.
--
--   begin
--       apex_application_install.set_authentication_scheme (
--           p_name => 'SSO-Production' );
--   end;
--   /
--   @f100.sql
--   begin
--       apex_application_install.set_authentication_scheme (
--           p_name => null );
--   end;
--   /
--   @f101.sql
--
-- See Also:
-- GET_AUTHENTICATION_SCHEME
--==============================================================================
procedure set_authentication_scheme (
    p_name in varchar2 );
--
--==============================================================================
-- Use this function to retrieve the authentication scheme name that should
-- override the default.
--
-- Example:
--
-- Print the authentication scheme override
--
--   select apex_application_install.get_authentication_scheme from sys.dual
--
-- See Also:
-- SET_AUTHENTICATION_SCHEME
--==============================================================================
function get_authentication_scheme
    return varchar2;
--
--==============================================================================
-- Use this procedure to set the REST Source Catalog group which a new REST
-- Source Catalog should be imported into.
--
-- Parameters
-- * p_name: The name of the REST Source Catalog Group. That Group must
--           exist in the workspace.
--
-- Example
--
-- Set the Catalog group to "Financial Services Catalogs". REST Source Catalogs will
-- be imported into this group.
--
--   begin
--       apex_application_install.set_rest_source_catalog_group (
--           p_name => 'Financial Services Catalogs' );
--   end;
--   /
--   @rest-service-catalog-financial.sql
--
-- See Also:
-- GET_REST_SOURCE_CATALOG_GROUP
--==============================================================================
procedure set_rest_source_catalog_group (
    p_name in varchar2 );
--
--==============================================================================
-- Use this function to retrieve the name of REST Source Catalog Group which
-- new catalogs will be imported into.
--
-- Example:
--
-- Print the REST Source Catalog Group override
--
--   select apex_application_install.get_rest_source_catalog_group from sys.dual
--
-- See Also:
-- SET_REST_SOURCE_CATALOG_GROUP
--==============================================================================
function get_rest_source_catalog_group
    return varchar2;
--
--==============================================================================
-- Use this function to override the build status for applications that are about
-- to be installed.
--
-- Parameters:
-- * p_build_status: New build status to set application to. Values include:
--                   - `RUN_AND_BUILD` - Developers and users can both run
--                     develop the application.
--                   - `RUN_ONLY` - Users can only run the application.
--                     Developers cannot edit the application.
--
-- Example:
--
-- The following example sets build status for app 100 to `RUN_ONLY`.
--
--   begin
--       apex_application_install.set_build_status (
--           p_build_status => 'RUN_ONLY' );
--   end;
--   /
--   @f100.sql
--
-- See Also:
-- GET_BUILD_STATUS
--==============================================================================
procedure set_build_status (
    p_build_status in varchar2 );
--
--==============================================================================
-- Use this function to retrieve the build status that should override the default.
--
-- Example:
--
-- The following example prints the build status override.
--
--   select apex_application_install.get_build_status from sys.dual
--
-- See Also:
-- SET_BUILD_STATUS
--==============================================================================
function get_build_status
    return varchar2;
--
--==============================================================================
-- This procedure clears all values currently maintained in the
-- APEX_APPLICATION_INSTALL package.
--
-- Example:
--
-- The following example clears all values currently set by the
-- APEX_APPLICATION_INSTALL package.
--
--   begin
--       apex_application_install.clear_all;
--   end;
--==============================================================================
procedure clear_all;
--
--==============================================================================
-- Use this function to retrieve install information from a source file.
--
-- Parameters:
-- * p_source: The source code, a table of (name, contents) with a single record
--             for normal APEX applications or multiple records for applications
--             that were split when exporting. Note that passing multiple
--             applications is not supported.
--
-- Returns:
--
-- This function returns information about the application that can be used to 
-- configure installation.
--
-- Raises:
-- * WWV_FLOW_IMP_PARSER.RUN_STMT_ERROR: The source contains invalid statements.
--
-- Example:
--
-- The following example fetches an application from a remote URL and prints its 
-- install information.
--
--   declare
--       l_source apex_t_export_files;
--       l_info   apex_application_install.t_file_info;
--   begin
--       l_source := apex_t_export_files (
--                       apex_t_export_file (
--                           name     => 'f100.sql',
--                           contents => apex_web_service.make_rest_request (
--                                           p_url         => 'https://www.example.com/apps/f100.sql',
--                                           p_http_method => 'GET' )));
--       l_info   := apex_application_install.get_info (
--                       p_source => l_source );
--
--       sys.dbms_output.put_line (
--           apex_string.format (
--               p_message => q'~
--                            !Type ................. %0
--                            !Workspace ............ %1
--                            !Version .............. %2
--                            !App ID ............... %3
--                            !App Name ............. %4
--                            !@Alias ................ %5
--                            !Owner ................ %6
--                            !Build Status ......... %7
--                            !Has Install Script ... %8
--                            !App ID Usage ......... %9
--                            !App Alias Usage ...... %10
--                            !~',
--               p0        => l_info.file_type,
--               p1        => l_info.workspace_id,
--               p2        => l_info.version,
--               p3        => l_info.app_id,
--               p4        => l_info.app_name,
--               p5        => l_info.app_alias,
--               p6        => l_info.app_owner,
--               p7        => l_info.build_status,
--               p8        => apex_debug.tochar(l_info.has_install_script),
--               p9        => l_info.app_id_usage,
--               p10       => l_info.app_alias_usage,
--               p_prefix  => '!' ));
--   end;
--
-- See Also:
-- - INSTALL
-- - GET_APPLICATION
--==============================================================================
function get_info (
    p_source in wwv_flow_t_export_files )
    return t_file_info;
--
--==============================================================================
-- Use this procedure to install an application. Use the
-- APEX_APPLICATION_INSTALL_INSTALL.SET* procedures to configure installation parameters.
--
-- Parameters:
-- * p_source:             The source code, a table of (name, contents) with
--                         a single record for normal Oracle Application Express
--                         applications or multiple records for applications that
--                         were split when exporting. Note that passing multiple
--                         applications is not supported. If NULL (the default),
--                         import the source that was previously passed to GET_INFO.
-- * p_overwrite_existing: If FALSE (the default), raise an error instead of 
--                         overwriting an existing application.
--
-- Raises:
-- * WWV_FLOW_IMP_PARSER.RUN_STMT_ERROR:                The source contains invalid statements.
-- * SECURITY_GROUP_ID_INVALID:                         The current workspace conflicts with the install
--                                                      workspace.
-- * WWV_FLOW_API.FLOW_ID_RESERVED_FOR_OTHER_WORKSPACE: The application ID isused in another workspace.
-- * WWV_FLOW_API.FLOW_ID_RANGE_RESERVED:               The application ID is reserved internal us.
-- * WWV_FLOW_API.FLOW_ID_OUT_OF_RANGE:                 The application ID used for installing is
--                                                      not in a valid range.
-- * APPLICATION_ID_RESERVED:                           The application ID is in use in the current workspace
--                                                      and p_overwrite_existing was set to false.
--
-- Example:
--
-- Fetch an application from a remote URL, then install it with a new ID
-- and new component ID offsets in workspace `EXAMPLE`.
--
--   declare
--       l_source apex_t_export_files;
--       l_info   apex_application_install.t_file_info;
--   begin
--       l_source := apex_t_export_files (
--           apex_t_export_file (
--               name     => 'f100.sql',
--               contents => apex_web_service.make_rest_request (
--                   p_url         => 'https://www.example.com/apps/f100.sql',
--                   p_http_method => 'GET' )));
--
--       apex_util.set_workspace('EXAMPLE');
--       apex_application_install.generate_application_id;
--       apex_application_install.generate_offset;
--       apex_application_install.install (
--           p_source => l_source );
--   end;
--
-- See Also:
-- - GET_INFO
-- - GET_APPLICATION
--==============================================================================
procedure install (
    p_source             in wwv_flow_t_export_files default null,
    p_overwrite_existing in boolean                 default false );
--
--==============================================================================
-- This procedure removes an application from a workspace. Use SET_WORKSPACE_ID or
-- SET_WORKSPACE to set the workspace.
--
-- Parameters
-- * p_application_id: The ID of the application.
--
-- Raises:
-- * WWV_FLOW_API.DELETE_APP_IN_DIFFERENT_WORKSPACE: The application is not in this workspace.
-- * WWV_FLOW_API.FLOW_NOT_DELETED:                  The application was not deleted.
-- * WWV_FLOW.APP_NOT_FOUND_ERR:                     The application ID was not found.
--
-- Example:
--
-- The following example demonstrates how to use the REMOVE_APPLICATION procedure
-- to remove an application with an ID of 100 from a workspace.
--
--   begin
--       apex_application_install.set_workspace('EXAMPLE');
--       apex_application_install.set_keep_sessions(false);
--       apex_application_install.remove_application(100);
--   end;
--
-- See Also:
-- - SET_WORKSPACE_ID
-- - SET_WORKSPACE
-- - SET_KEEP_SESSIONS
--==============================================================================
procedure remove_application (
    p_application_id in number );
--
--##############################################################################
-- Translation
--##############################################################################
--
--==============================================================================
-- Internal:
-- Publish Application
--==============================================================================
procedure publish_application (
    p_application_id        in number,
    p_language_code         in varchar2,
    p_new_translated_app_id in number   default null );

end wwv_flow_application_install;
/
show errors
