/*!
 * Copyright (c) 2013, 2021, Oracle and/or its affiliates. All rights reserved.
 */

/**
 * @fileOverview
 * Turns a HTML textarea into a code editor for css, javascript, html and pl/sql
 *   apex.jQuery( "#myEditor" ).codeEditor({...});
 */

/* global monaco,require */
/* eslint-disable no-bitwise*/

(function ($, util, lang, locale, debug, actions, server, env) {
    'use strict';

    const msg = key => lang.getMessage('CODE_EDITOR.' + key);

    // constants
    const C_ACTIVE = 'is-active',
        C_CODE_EDITOR = 'a-MonacoEditor',
        C_CONTENT_WRAPPER = 'a-MonacoEditorContent',
        C_TOOLBAR = 'a-MonacoEditor-toolbar',
        SEL_TOOLBAR = '.' + C_TOOLBAR,
        C_NOTIFICATION = 'a-MonacoEditor-notification',
        SEL_NOTIFICATION = '.' + C_NOTIFICATION,
        C_NOTIFICATION_MSG = 'a-MonacoEditor-message',
        SEL_NOTIFICATION_MSG = '.' + C_NOTIFICATION_MSG,
        C_ACTUAL_EDITOR = 'a-MonacoEditor-editor';

    const LANG_JAVASCRIPT = 'javascript',
        LANG_SQL = 'sql',
        LANG_MLE_JAVASCRIPT = 'mle-javascript',
        LANG_QUICKSQL = 'quicksql';

    // text messages
    const MSG_ACCESSIBILITY_MODE =        msg('ACCESSIBILITY_MODE'),
        MSG_CLOSE =                     msg('CLOSE'),
        MSG_FIND =                      msg('FIND'),
        MSG_HINT =                      msg('HINT'),
        MSG_INDENTATION =               msg('INDENTATION'),
        MSG_INDENT_SIZE =               msg('INDENT_SIZE'),
        MSG_INDENT_WITH_TABS =          msg('INDENT_WITH_TABS'),
        //MSG_ITEM_DOES_NOT_EXIST =     msg('ITEM_DOES_NOT_EXIST'),
        MSG_MINIMAP =                   msg('MINIMAP'),
        MSG_MINIMAP_CONDITIONAL =       msg('MINIMAP_CONDITIONAL'),
        MSG_QUERY_BUILDER =             msg('QUERY_BUILDER'),
        MSG_REDO =                      msg('REDO'),
        MSG_SETTINGS =                  msg('SETTINGS'),
        MSG_SHOW_LINE_NUMBERS =         msg('SHOW_LINE_NUMBERS'),
        MSG_SHOW_RULER =                msg('SHOW_RULER'),
        MSG_SUGGESTIONS =               msg('SUGGESTIONS'),
        MSG_TAB_SIZE =                  msg('TAB_SIZE'),
        MSG_THEMES =                    msg('THEMES'),
        MSG_THEME_AUTOMATIC =           msg('THEME_AUTOMATIC'),
        MSG_THEME_DARK =                msg('THEME_DARK'),
        MSG_THEME_HIGH_CONTRAST_DARK =  msg('THEME_HIGH_CONTRAST_DARK'),
        MSG_THEME_LIGHT =               msg('THEME_LIGHT'),
        MSG_TRANSFORM_CODE_CASE =       msg('TRANSFORM_CODE_CASE'),
        MSG_UNDO =                      msg('UNDO'),
        MSG_UTIL_OFF =                  msg('UTIL_OFF'),
        MSG_UTIL_ON =                   msg('UTIL_ON'),
        MSG_VALIDATE =                  msg('VALIDATE'),
        MSG_VALIDATION_SUCCESS =        msg('VALIDATION_SUCCESS'),
        MSG_WHITESPACE =                msg('WHITESPACE');

    const PREFERENCE_KEYS = [
            'theme', 'tabsInsertSpaces', 'indentSize',
            'tabSize', 'ruler', 'minimap', 'accessibilityMode',
            'lineNumbers', 'whitespace', 'showSuggestions'
        ],
        OPTIONS_THEME = [
            //automatic will be either vs or vs-dark depending on the builder setting
            { label: MSG_THEME_AUTOMATIC, value: 'automatic' },
            { label: MSG_THEME_LIGHT, value: 'vs' },
            { label: MSG_THEME_DARK, value: 'vs-dark' },
            { label: MSG_THEME_HIGH_CONTRAST_DARK, value: 'hc-black' }
        ],
        MINIMAP_OFF = 'off',
        MINIMAP_ON = 'on',
        MINIMAP_CONDITIONAL = 'conditional',
        OPTIONS_MINIMAP = [
            { label: MSG_UTIL_ON, value: MINIMAP_ON },
            { label: MSG_UTIL_OFF, value: MINIMAP_OFF },
            { label: MSG_MINIMAP_CONDITIONAL, value: MINIMAP_CONDITIONAL },
        ],
        OPTIONS_TAB_SIZE = [
            { label: '2', value: '2' },
            { label: '3', value: '3' },
            { label: '4', value: '4' },
            { label: '8', value: '8' }
        ],
        OPTIONS_INDENT_SIZE = [
            { label: '2', value: '2' },
            { label: '3', value: '3' },
            { label: '4', value: '4' },
            { label: '8', value: '8' }
        ];

    // absolute paths are needed by the webworkers
    // if the image directory starts with '/' or '.', it is a relative path which must be made absolute
    const ABSOLUTE_PATH =
            ['/', '.'].includes( env.APEX_FILES[0] )
                ? document.location.protocol + '//' + document.location.host + (env.APEX_FILES[0] === '.' ? '/' : '') + env.APEX_FILES
                : env.APEX_FILES,
        SOURCE_PATH = ABSOLUTE_PATH + 'libraries/monaco-editor/0.22.3/min',
        APEX_JS_DTS_PATH = ABSOLUTE_PATH + 'libraries/monaco-editor/apex/apex-js-api.d.ts?v=' + env.APEX_VERSION,
        MLE_DTS_PATH = ABSOLUTE_PATH + 'libraries/monaco-editor/apex/mle-js-oracledb.d.ts?v=' + env.APEX_VERSION,
        QUICKSQL_PATH = ABSOLUTE_PATH + 'libraries/monaco-editor/apex/quicksql.js?v=' + env.APEX_VERSION;

    // Disposable keys
    const DISP_DTS_APEX = 'dts-apex',
        DISP_DTS_APEX_ITEMS_REGIONS = 'dts-apex-items-regions',
        DISP_DTS_APEX_MLE = 'dts-mle',
        DISP_MLE_ENV = 'mle-env',
        DISP_MINIMAP = 'minimap';
    
    // everything else will be defaulted to the base "item" or "region" type
    // should be kept in sync with the currently documented item/region types
    const DTS_TYPES = {
        ITEM: {
            NATIVE_NUMBER_FIELD: "numberFieldItem",
            NATIVE_COLOR_PICKER: "colorPickerItem"
        },
        REGION: {
            NATIVE_FACETED_SEARCH: "facetsRegion",
            NATIVE_SMART_FILTERS: "facetsRegion",
            NATIVE_MAP_REGION: "mapRegion"
        }
    };

    let currentTheme;

    // keeping track of various disposables returned by monaco functions
    // in case we need to dispose (revert) them later
    const disposableManager = (function(){
        var DISPOSABLES = {
            GLOBAL: {} // for disposables shared across all instances, eg. javascript language defaults
            // other keys will be added as editors are created
        };

        return {
            // global disposables
            existsGlobally: key => {
                return DISPOSABLES.GLOBAL[ key ] !== undefined;
            },
            registerGlobally: ( key, disposable ) => {
                if ( DISPOSABLES.GLOBAL[ key ] ) {
                    DISPOSABLES.GLOBAL[ key ].dispose();
                }
                DISPOSABLES.GLOBAL[ key ] = disposable;
            },
            disposeGlobally: key => {
                if (DISPOSABLES.GLOBAL[ key ]) {
                    DISPOSABLES.GLOBAL[ key ].dispose();
                    delete DISPOSABLES.GLOBAL[ key ];
                }
            },
            // editor specific disposables
            deregisterEditor: ( editorId ) => {
                if(DISPOSABLES[ editorId ]){
                    for (var key of Object.keys(DISPOSABLES[ editorId ])) {
                        DISPOSABLES[ editorId ][ key ].dispose();
                        delete DISPOSABLES[ editorId ][key];
                    }
                    delete DISPOSABLES[ editorId ];
                }
            },
            existsForEditor: ( editorId, key ) => {
                return DISPOSABLES[ editorId ] && DISPOSABLES[ editorId ][ key ] !== undefined;
            },
            registerForEditor: ( editorId, key, disposable ) => {
                DISPOSABLES[ editorId ] = DISPOSABLES[ editorId ] || {};
                if ( DISPOSABLES[ editorId ][ key ] ) {
                    DISPOSABLES[ editorId ][ key ].dispose();
                }
                DISPOSABLES[ editorId ][ key ] = disposable;
            },
            disposeForEditor: ( editorId, key ) => {
                if(DISPOSABLES[ editorId ] && DISPOSABLES[ editorId ][ key ]){
                    DISPOSABLES[ editorId ][ key ].dispose();
                    delete DISPOSABLES[ editorId ][ key ];
                }
            }
        };
    })();
    function loadExtraDtsFromPath(key, path) {
        fetch(path)
            .then(response => response.text())
            .then(text => {
                disposableManager.registerGlobally(key, monaco.languages.typescript.javascriptDefaults.addExtraLib(text));
            })
            .catch(error => debug.error('Code Editor: could not fetch Typescript declaration file from ' + path, error));
    }
    function loadExtraDtsFromString(key, string) {
        disposableManager.registerGlobally(key, monaco.languages.typescript.javascriptDefaults.addExtraLib(string));
    }
    function loadJsApiDts( pageInfo ) {
        // the main dts is only loaded once
        if (!disposableManager.existsGlobally( DISP_DTS_APEX )) {
            loadExtraDtsFromPath(DISP_DTS_APEX, APEX_JS_DTS_PATH);
        }
        // the items and regions dts snippet is disposed and generated each time an editor is created
        // just in case the page items/regions have changed
        // Note: apex.items/regions autocomplete *should* be registered as *editor specific* disposables
        //  and they should only be applied to a specific editor instance. however, due to the mechanism used
        //  (addExtraLib) this can only be done globally
        disposableManager.disposeGlobally( DISP_DTS_APEX_ITEMS_REGIONS );
        loadExtraDtsFromString(DISP_DTS_APEX_ITEMS_REGIONS, `
            declare namespace apex {
                var items: {
                    ${ pageInfo.items.map(item => `${ JSON.stringify(item.name) }: ${DTS_TYPES.ITEM[item.type] || "item"};`).join("\n") }
                };
                var regions: {
                    ${ pageInfo.regions.map(region => `${ JSON.stringify(region.staticId) }: ${DTS_TYPES.REGION[region.type] || "region"};`).join("\n") }
                };
            }
        `);
    }
    function unloadJsApiDts() {
        disposableManager.disposeGlobally(DISP_DTS_APEX);
        disposableManager.disposeGlobally(DISP_DTS_APEX_ITEMS_REGIONS);
    }
    function loadMleDts() {
        if (!disposableManager.existsGlobally( DISP_DTS_APEX_MLE )) {
            loadExtraDtsFromPath(DISP_DTS_APEX_MLE, MLE_DTS_PATH);
        }
    }
    function unloadMleDts() {
        disposableManager.disposeGlobally(DISP_DTS_APEX_MLE);
    }

    function loadLanguageDependencies(language, pageInfo) {
        return new Promise(resolve => {
            if([LANG_JAVASCRIPT, LANG_MLE_JAVASCRIPT].includes( language )){
                // this function should be called when loading or updating the language to javascript or mle-javascript
                monaco.languages.typescript.javascriptDefaults.setCompilerOptions({
                    target: monaco.languages.typescript.ScriptTarget.ES6,
                    allowJs: true,
                    allowNonTsExtensions: true,
                    // in mle mode, browser globals like window or console should be hidden
                    noLib: language === LANG_MLE_JAVASCRIPT
                });
        
                monaco.languages.typescript.javascriptDefaults.setDiagnosticsOptions({
                    // do not mark nameless functions as errors
                    // needed for Init JS Code
                    diagnosticCodesToIgnore: [1003]
                });
        
                if (language === LANG_JAVASCRIPT) {
                    unloadMleDts();
                    loadJsApiDts(pageInfo);
                } else if (language === LANG_MLE_JAVASCRIPT) {
                    unloadJsApiDts();
                    loadMleDts();
                }
                resolve();
            } else if ( language === LANG_QUICKSQL ){
                if (!monaco.languages.getEncodedLanguageId(LANG_QUICKSQL)) {
                    server.loadScript({
                        path: QUICKSQL_PATH
                    }, resolve);
                } else {
                    resolve();
                }
            } else {
                resolve();
            }
        });
    }

    // should be called once, before monaco was loaded, but after requirejs was loaded
    function configureMonacoPreLoad() {
        // configuring monaco
        // https://github.com/microsoft/monaco-editor/blob/master/docs/integrate-amd-cross.md
        // Before loading vs/editor/editor.main, define a global MonacoEnvironment that overwrites
        // the default worker url location (used when creating WebWorkers). The problem here is that
        // HTML5 does not allow cross-domain web workers, so we need to proxy the instantiation of
        // a web worker through a same-domain script
        window.MonacoEnvironment = {
            getWorkerUrl: function () {
                return 'data:text/javascript;charset=utf-8,' + encodeURIComponent([
                    'self.MonacoEnvironment = {',
                    `    baseUrl: "${SOURCE_PATH}"`,
                    '};',
                    `importScripts("${SOURCE_PATH}/vs/base/worker/workerMain.js");`
                ].join('\n'));
            }
        };
        require.config({
            paths: { 'vs': SOURCE_PATH + '/vs' },
            waitSeconds: 0
        });
        // monaco supports en, de, es, fr, it, ja, ko, ru, zh-tw and zh-cn
        // if none is matched by the current locale, we default to english
        var browserLang = locale.getLanguage().toLowerCase();
        var monacoLang;
        // only ones for which we need to take into account the country
        if (['zh-tw', 'zh-cn'].includes(browserLang)) {
            monacoLang = browserLang;
        } else {
            // slicing off a possible country portion
            browserLang = browserLang.split('-')[0];
            if (['de', 'es', 'fr', 'it', 'ja', 'ko', 'ru'].includes(browserLang)) {
                monacoLang = browserLang;
            }
        }
        // if a special language was matched, we set it. otherwise english stays
        if (monacoLang) {
            require.config({
                'vs/nls': {
                    availableLanguages: {
                        '*': monacoLang
                    }
                }
            });
        }
    }

    // should be called once, after monaco was loaded
    function configureMonacoPostLoad() {
        // this is the central place to alter any default monaco theme colors
        monaco.editor.defineTheme('apex-vs', {
            inherit: true,
            base: 'vs',
            rules: [
                // background color: fffffe
                // sql
                { token: 'string.sql', foreground: 'b26100' },          // 4.56  AA
                { token: 'keyword.sql', foreground: 'c74634' },         // 4.82  AA
                { token: 'predefined.sql', foreground: '7e5e8a' },      // 5.44  AA
                { token: 'operator.sql', foreground: '000000' },        // 20.99 AAA
                { token: 'atom.sql', foreground: '398459' },            // 4.54  AA
                { token: 'function.sql', foreground: '795E26' },        // 6.10  AA
                { token: 'pageitem.sql', foreground: '008080', fontStyle: 'bold' },
                // quicksql
                { token: 'table.quicksql', foreground: '1E84BF', fontStyle: 'bold' },   // 4.11 AA Large
                { token: 'view.quicksql', foreground: '008855', fontStyle: 'bold' },    // 4.51 AA
                { token: 'list.quicksql', foreground: 'b26100' },                       // 4.56 AA
                { token: 'keywords.quicksql', foreground: 'c74634' },                   // 4.82 AA
                { token: 'types.quicksql', foreground: '398459' },                      // 4.54 AA
            ]
        });
        monaco.editor.defineTheme('apex-vs-dark', {
            inherit: true,
            base: 'vs-dark',
            rules: [
                // background color: 1e1e1e
                // sql
                { token: 'string.sql', foreground: 'ecbb76' },          // 9.48  AAA
                { token: 'keyword.sql', foreground: 'f14840' },         // 4.56  AA
                { token: 'predefined.sql', foreground: 'a687b3' },      // 5.35  AA
                { token: 'operator.sql', foreground: 'D4D4D4' },        // 11.25 AAA
                { token: 'atom.sql', foreground: '259856' },            // 4.53  AA
                { token: 'function.sql', foreground: 'DCDCAA' },        // 11.80 AAA
                { token: 'pageitem.sql', foreground: '3dc9b0', fontStyle: 'bold' },
                // quicksql
                { token: 'table.quicksql', foreground: '1E84BF', fontStyle: 'bold' },   // 4.06 AA   Large
                { token: 'view.quicksql', foreground: '259856', fontStyle: 'bold' },    // 4.53 AA
                { token: 'list.quicksql', foreground: 'ecbb76' },                       // 9.48 AAA
                { token: 'keywords.quicksql', foreground: 'f14840' },                   // 4.56 AA
                { token: 'types.quicksql', foreground: 'A3CDFF' },                      // 4.53 AA
            ]
        });
        monaco.editor.defineTheme('apex-hc-black', {
            inherit: true,
            base: 'hc-black',
            rules: [
                // background color: 000000
                // sql
                { token: 'string.sql', foreground: 'ecbb76' },          // 11.94 AAA
                { token: 'keyword.sql', foreground: 'f14840' },         // 5.74  AA
                { token: 'predefined.sql', foreground: 'a687b3' },      // 6.74  AA
                { token: 'operator.sql', foreground: 'D4D4D4' },        // 14.17 AAA
                { token: 'atom.sql', foreground: '259856' },            // 5.70  AA
                { token: 'function.sql', foreground: 'DCDCAA' },        // 14.86 AAA
                { token: 'pageitem.sql', foreground: '3dc9b0', fontStyle: 'bold' },
                // quicksql
                { token: 'table.quicksql', foreground: '1E84BF', fontStyle: 'bold' },   // 5.11  AA
                { token: 'view.quicksql', foreground: '259856', fontStyle: 'bold' },    // 5.73  AA
                { token: 'list.quicksql', foreground: 'ecbb76' },                       // 11.94 AAA
                { token: 'keywords.quicksql', foreground: 'f14840' },                   // 5.74  AA
                { token: 'types.quicksql', foreground: '259856' },                      // 5.73  AA
            ]
        });
    }

    // resolves when all dependencies have been loaded
    // 1) requirejs, can already be present on the page, if not, it will load async
    // 2) monaco, can already be present on the page, if not, it will load async
    const loadEditorDependencies = new Promise( mainResolve => {
        new Promise( requirejsresolve => {
            if (window.require === undefined) {
                server.loadScript({
                    path: SOURCE_PATH + '/vs/loader.js'
                }, requirejsresolve);
            } else {
                requirejsresolve();
            }
        })
        .then( configureMonacoPreLoad )
        .then( () => {
            // no matter if the editor script tag is already on the page or not
            // it still has to be loaded via require
            require(['vs/editor/editor.main'], () => {
                configureMonacoPostLoad();
                mainResolve();
            });
        });
    });

    // returns the external theme preference: vs or vs-dark
    function getThemeForAutomatic() {
        if( apex.builder ){
            // if in the builder, respect the builder theme, which might or might not be based on the OS preference
            return apex.builder.isBuilderDarkMode() ? 'vs-dark' : 'vs';
        } else {
            // if outside of the builder, respect the OS preference
            return window.matchMedia && window.matchMedia( '(prefers-color-scheme: dark)' ).matches ? 'vs-dark' : 'vs';
        }
    }

    // sets the monaco theme
    // this can only be done globally
    // pass in any of the valid themes: automatic, vs, vs-dark or hc-black
    function setTheme( theme ) {

        currentTheme = theme;

        let themeToSet;

        if (theme === 'automatic') {
            themeToSet = getThemeForAutomatic();
        } else if (['vs', 'vs-dark', 'hc-black'].includes(theme)) {
            themeToSet = theme;
        } else {
            themeToSet = getThemeForAutomatic();
            debug.info(theme + ' is not a valid Monaco theme. Using automatic instead.');
        }

        // ensures we only touch monaco once it's actually loaded
        loadEditorDependencies.then( () => {
            monaco.editor.setTheme( "apex-" + themeToSet );
        });
    }

    // when running in the APEX builder, theme automatic is superseded by the builder theme
    //  which can itself be automatic
    // when running outside of the builder, theme automatic comes straight from the OS
    //
    // in both cases, on theme change, if the current editor theme is automatic, we reevaluate it
    if( apex.builder ){
        // triggered by builder.js on theme change
        $( "body" ).on( "apex-builder-theme-changed", () => {
            if( currentTheme === "automatic" ){
                setTheme( "automatic" );
            }
        } );
    } else {
        if ( window.matchMedia ){
            window.matchMedia( "(prefers-color-scheme: dark)" ).addListener( () => {
                if( currentTheme === "automatic" ){
                    setTheme( "automatic" );
                }
            });
        }
    }

    function isMac() {
        return /(Mac|iPhone|iPod|iPad)/i.test(navigator.platform);
    }

    function getUniversalCtrlKey() {
        // WinCtrl = Ctrl on Mac, WinKey on Windows
        // CtrlCmd = Cmd on Mac, Ctrl on Windows
        // this conditional will ensure Ctrl on both
        return isMac() ? monaco.KeyMod.WinCtrl : monaco.KeyMod.CtrlCmd;
    }

    // Tries to lowercase a string. If the string is already lower-cased, it will uppercase it
    // parts of the string contained enclosed in ' ', " " (strings or case sensitive aliases) are *not* transformed
    function transformCodeCase(str) {

        if (!str) {
            return str;
        }

        function getRangesToIgnore(str) {
            var arr = [];
            var matches = str.matchAll(/('|")[\s\S]*?(\1)/gm);
            for (var match of matches) {
                arr.push({ start: match.index, end: match.index + match[0].length, ignore: true });
            }
            return arr;
        }

        function completeRanges(ranges, lastIndex) {
            var finalRanges = [];
            var currentIndex = 0;

            function addRange(start, end, ignore) {
                finalRanges.push({ start: start, end: end, ignore: ignore });
            }

            if (ranges.length === 0) {
                addRange(0, lastIndex, false);
            } else {
                for (var i = 0; i < ranges.length; i++) {
                    var range = ranges[i];
                    if (currentIndex < range.start) {
                        addRange(currentIndex, range.start - 1, false);
                    }
                    addRange(range.start, range.end, true);
                    currentIndex = range.end + 1;
                }
                if (currentIndex < lastIndex - 1) {
                    addRange(currentIndex, lastIndex, false);
                }

            }

            return finalRanges;
        }

        function applyFunction(ranges, func) {
            var i, range, strPart, result = '';
            for (i = 0; i < allRanges.length; i++) {
                range = allRanges[i];
                strPart = str.slice(range.start, range.end + 1);
                result += range.ignore ? strPart : func(strPart);
            }
            return result;
        }

        function toLowerCase(ranges) {
            return applyFunction(ranges, function (str) {
                return str.toLocaleLowerCase();
            });
        }

        function toUpperCase(ranges) {
            return applyFunction(ranges, function (str) {
                return str.toLocaleUpperCase();
            });
        }

        var rangesToIgnore = getRangesToIgnore(str);
        var allRanges = completeRanges(rangesToIgnore, str.length - 1);

        var strLowerCase = toLowerCase(allRanges);
        if (str !== strLowerCase) {
            return strLowerCase;
        } else {
            return toUpperCase(allRanges);
        }
    }

    $.widget('apex.codeEditor', {
        version: '20.2',
        widgetEventPrefix: 'codeEditor',
        options: {
            language: LANG_JAVASCRIPT,
            value: '',
            readOnly: false,
            autofocus: false,
            ariaLabel: '',
            errors: [],
            warnings: [],
            /*
             * Editor specific suggestions
             * Can be an array of objects:
             *  [{
             *      label: 'P1_FIRST_NAME (First Name)', // required
             *      insertText: 'P1_FIRST_NAME',         // optional
             *      detail: 'Page Item',                 // optional
             *      documentation: 'some  text'          // optional
             *  }]
             *
             * Or a callback function that will be invoked on widget load and on language change
             * It must return the same kind of array
             * function(language){
             *      if(language == 'sql'){ return [...];}
             *      else if(language == 'javascript'){ return [...];}
             *      else { return []; }
             * }
             */
            suggestions: null,
            /*
             * pageInfo should provided only if in Page Designer
             * {
             *    pageId: 1,
             *    items: [{
             *      name: "P1_ID",
             *      type: "NATIVE_NUMBER_FIELD"
             *    }]
             *    regions: [{
             *      staitcId: "myFacetRegion",
             *      type: "NATIVE_FACETED_SEARCH"
             *    }]
             * }
             * 
             * pageId is used by the item name validator
             * itemNames is used by the item name validator and by the apex.items autocomplete
             * regionStaticIds is used by the apex.regions autocomplete
             */
            pageInfo: {
                pageId: null,
                items: [],
                regions: []
            },
            showSuggestions: true,
            // should be passed as true in Page Designer, while also assuring the page items are returned by the suggestions function
            validatePageItems: false,
            accessibilityMode: false,
            // monaco options: exposing them here as opposed to arbitrarily passing passing them forward
            // to keep a list of all options used

            // "on", "off", "conditional" (only show when content >= 100 lines)
            minimap: MINIMAP_CONDITIONAL,
            lineNumbers: true,
            wordWrap: false,
            scrollBeyondLastLine: true,
            // wordBasedSuggestions are not scope-limited, so they can be more disturbing than useful
            wordBasedSuggestions: false,
            // toolbar is always available in the builder, but can be disabled for theme roller
            toolbar: true,
            whitespace: false,
            // settings which can be overridden by user preference
            theme: 'automatic',
            tabsInsertSpaces: true,
            indentSize: '4',
            tabSize: '4',
            ruler: false,
            // callback functions
            onInitialized: null,  // optional. function that runs after the editor is initialized. function(editor){}
            codeComplete: null,   // optional. function( options, callback )
            validateCode: null,   // optional. function( code, callback ) callback: function( {errors:[],warnings:[]} )
            queryBuilder: null,   // optional. function( editor, code )
            heightFn: null,       // optional. function returning editor height
                                  // if provided, it will be called on initialization and subsequently at window resize
            // events/callbacks
            preferencesChanged: null    // function( event )
        },
        // holds monaco context keys which can be used as conditions in a number of places, eg. keyboard shortcuts
        _contextKeys: {},
        /*
         * Lifecycle methods
         */
        _create: function() {
            const self = this,
                o = this.options,
                parentElement = self.element[0],
                parentElement$ = $(parentElement),
                contentWrapper$ = $(`<div class="${C_CONTENT_WRAPPER}"></div>`),
                toolbar$ = $(`<div class="${C_TOOLBAR}"></div>`),
                editor$ = $(`<div class="${C_ACTUAL_EDITOR}"></div>`);

            self.element.addClass(C_CODE_EDITOR);
            self.baseId = self.element[0].id;

            parentElement$.append(contentWrapper$);

            if (o.toolbar) {
                contentWrapper$.append(toolbar$);
                self._initNotificationBar();
                self._context = self._createContext();
                self._initToolbar(toolbar$);
            }

            contentWrapper$.append(editor$);

            ( async () => {
                await loadEditorDependencies;
                await loadLanguageDependencies(this.options.language, this.options.pageInfo);
    
                // monaco themes are set globally
                setTheme( o.theme );
    
                var editor = monaco.editor.create(editor$[0], {
                    value: o.value,
                    language: o.language.replace('mle-', ''),
                    readOnly: o.readOnly,
                    detectIndentation: false,
                    insertSpaces: o.tabsInsertSpaces,
                    indentSize: o.indentSize,
                    tabSize: o.tabSize,
                    rulers: o.ruler ? [80] : [],
                    scrollBeyondLastLine: o.scrollBeyondLastLine,
                    lineNumbers: o.lineNumbers,
                    // slightly narrower line number column
                    lineNumbersMinChars: 4,
                    scrollbar: {
                        // allows page to scroll if bottom of editor was reached
                        alwaysConsumeMouseWheel: false
                    },
                    fixedOverflowWidgets: true,
                    ariaLabel: o.ariaLabel,
                    // when editing plaintext, enforce the wordWrap option
                    // otherwise respect the passed in options
                    wordWrap: o.language === 'plaintext' ? true : (o.wordWrap ? 'on' : 'off'),
                    wordBasedSuggestions: o.wordBasedSuggestions,
                    quickSuggestions: o.showSuggestions,
                    accessibilitySupport: o.accessibilityMode ? 'on' : 'off',
                    renderWhitespace: o.whitespace ? 'all' : 'none',
                    // content should be copied plaintext for clean pasting in rich text contexts (Word, Jira etc)
                    copyWithSyntaxHighlighting: false
                });
    
                self._editor = editor;
                self._editorId = editor.getId();
    
                // editor-specific disposableManager logic
                editor.onDidDispose(self._disposeAllDisposables);
    
                // initialize monaco context keys used as conditions for various actions
                self._evaluateContextKeys(o.language);
    
                if (o.toolbar) {
                    self._updateNotifications();
                    self._populateContext();
                }
    
                // resizing logic
                window.addEventListener('resize', util.debounce(function () {
                    self.resize();
                }, 200));
                parentElement$.on('resize', function (pEvent) {
                    self.resize();
                    pEvent.stopPropagation();
                });
    
                if (o.autofocus) {
                    self.focus();
                }
    
                // Use Monaco onDidChangeContent to simulate a keypress on text area, allowing DA 'Key press' to work
                editor.getModel().onDidChangeContent(function (event) {
                    self.element.children('textarea').trigger('keypress', event);
                });
    
                // Combine onDidFocusEditorWidget and onDidBlurEditorWidget to simulate change event
                var oldValue;
                editor.onDidFocusEditorWidget(function () {
                    oldValue = editor.getValue();
                });
    
                editor.onDidBlurEditorWidget(function () {
                    var textarea$ = self.element.children('textarea'),
                        newValue = editor.getValue();
                    // Trigger blur on the textarea, so DA 'blur' will work
                    textarea$.trigger('blur');
                    // Simulate 'change' event (on blur and value changed), so DA 'change' will work
                    if (oldValue !== newValue) {
                        textarea$.val(newValue).trigger('change');
                    }
                });
    
                // initialize item name validator
                if (o.validatePageItems && o.pageInfo.pageId) {
                    self._initializeItemValidator();
                }
    
                // editor specific suggestion provider based on o.suggestions and o.language
                self._configureCustomSuggestionProvider();
    
                // configures custom shortcuts and extra editor actions
                self._initializeCustomActions();
    
                self._setupMinimap();
    
                // the accessibility mode can also be changed via the command pallette
                // we want to ensure our widget/toolbar option stays in sync with this editor option
                var accessibilitySupportId = monaco.editor.EditorOptions.accessibilitySupport.id;
                editor.onDidChangeConfiguration(function(evt){
                    if(evt.hasChanged(accessibilitySupportId)){
                        o.accessibilityMode = (editor.getOption(accessibilitySupportId) === monaco.editor.AccessibilitySupport.Enabled);
                        self._notifyPreferenceChange();
                    }
                });
    
                // monaco uses a textarea deep inside the widget, onto which at times a change event is triggered
                // this can cause issues when embedding the editor in other widgets, such as Theme Roller
                // a fix is to not let any internal change events propagate outside of the editor
                editor$.on('change', function(e){
                    e.stopPropagation();
                });
    
                if(self._tempValue !== undefined){
                    editor.setValue(self._tempValue);
                    self._tempValue = undefined;
                }
    
                self._on(self._eventHandlers);
    
                if (o.onInitialized) {
                    o.onInitialized(editor);
                }
            } )();
        },

        _eventHandlers: {
            resize: function (event) {
                this.resize();
                event.stopPropagation();
            },
            focusin: function () {
                this.element.addClass(C_ACTIVE);
            },
            focusout: function () {
                this.element.removeClass(C_ACTIVE);
            }
        },

        _destroy: function () {
            if (this._editor) {
                this._editor.dispose();
            }
        },

        _setOption: function (key, value) {
            var self = this,
                o = self.options,
                editor = self._editor;
            self._super(key, value);
            if (['errors', 'warnings'].includes(key)) {
                self._updateNotifications();
            } else if (key === 'language') {
                self._setLanguage(value);
            } else if (key === 'tabsInsertSpaces') {
                editor.updateOptions({ insertSpaces: value });
            } else if (key === 'ruler') {
                editor.updateOptions({ rulers: value ? [80] : [] });
            } else if (key === 'theme') {
                o.theme = value;
                setTheme( value );
            } else if (key === 'minimap') {
                self._setupMinimap();
            } else if (key === 'lineNumbers') {
                editor.updateOptions({ lineNumbers: !!value });
            } else if (key === 'accessibilityMode') {
                editor.updateOptions({ accessibilitySupport: value ? 'on' : 'off' });
            } else if (key === 'whitespace'){
                editor.updateOptions({ renderWhitespace: value ? 'all' : 'none' });
            } else if (key === 'showSuggestions'){
                editor.updateOptions({ quickSuggestions: !!value });
            } else {
                editor.updateOptions({ key: value });
            }
        },

        setValue: function (value) {
            if(this._editor){
                this._editor.setValue(value);
            } else {
                this._tempValue = value;
            }
        },

        getValue: function () {
            if (this._editor) {
                return this._editor.getValue();
            } else if(this._tempValue !== undefined){
                // the editor has not yet initialized, but a new value has already been set
                return this._tempValue;
            } else {
                // the editor has not yet initialized. returning original initial value
                return this.options.value;
            }
        },

        getEditor: function () {
            return this._editor;
        },

        resize: function () {
            var self = this,
                editor = self._editor;

            if(self.options.heightFn){
                $('#' + self.baseId).outerHeight(self.options.heightFn());
            }

            if (editor) {
                editor.layout();
            }
        },

        getSelection: function () {
            return this._editor.getModel().getValueInRange(this._editor.getSelection());
        },

        /**
         * The caller is responsible for making sure that pMessage is escaped as needed.
         * @param pMessage may contain markup
         */
        showNotification: function (message) {
            var container$ = this.element.find(SEL_NOTIFICATION).show().children(SEL_NOTIFICATION_MSG).first();
            if (typeof message === 'string') {
                container$.html(message);
            } else if (typeof message === 'object' && message instanceof jQuery) {
                container$.empty();
                container$.append(message);
            } else {
                debug.error('The notification message must be a string or a jQuery object');
            }
            this.resize();
        },

        // removes the message content and hides the container
        resetNotification: function () {
            this.element.find(SEL_NOTIFICATION).children(SEL_NOTIFICATION_MSG).empty();
            this.element.find(SEL_NOTIFICATION).hide();
            this.resize();
        },

        // the message must be escaped externally
        showSuccessNotification: function (message) {
            this.showNotification(`<ul><li class="is-success">${message}</li></ul>`);
        },

        focus: function () {
            this.element.addClass(C_ACTIVE);
            this._editor.focus();
        },

        // note that lineNumber and column are 1-based indexes
        setCursor: function (lineNumber, column) {
            this._editor.setPosition({
                lineNumber: lineNumber,
                column: column
            });
        },

        setCursorToEnd: function (revealLine) {
            var editor = this._editor,
                model = editor.getModel(),
                lastLine = model.getLineCount(),
                lastColumn = model.getLineMaxColumn(lastLine);

            this.setCursor(lastLine, lastColumn);

            if (revealLine) {
                editor.revealLine(lastLine);
            }
        },

        changeGeneration: function () {
            return this._editor.getModel().getAlternativeVersionId();
        },

        isClean: function (pGeneration) {
            return this._editor.getModel().getAlternativeVersionId() === pGeneration;
        },

        getPreferencesString: function () {
            var o = this.options,
                obj = {};

            for (var i = 0; i < PREFERENCE_KEYS.length; i++) {
                var key = PREFERENCE_KEYS[i];
                obj[key] = o[key];
            }

            return JSON.stringify(obj);
        },

        /*
         * Private functions
         */

        // editor-specific disposable logic
        _disposableExists: function( key ){
            return disposableManager.existsForEditor( this._editorId, key );
        },
        _registerDisposable: function( key, disposable ){
            disposableManager.registerForEditor( this._editorId, key, disposable );
        },
        _disposeDisposable: function( key ){
            disposableManager.disposeForEditor( this._editorId, key );
        },
        _disposeAllDisposables: function(){
            disposableManager.deregisterEditor( this._editorId );
        },
        // triggers an internal monaco action
        _triggerMonacoAction: function (actionName) {
            const editor = this._editor;

            // the editor must be in focus for the action to take effect
            this.focus();
            editor.trigger(null, actionName);
        },

        // to be called after widget initializtion and when the language changes
        _configureCustomSuggestionProvider: function () {
            var self = this,
                editor = self._editor,
                o = self.options,
                language = o.language.replace('mle-', ''),
                suggestions = o.suggestions,
                finalSuggestions,
                word;

            if (suggestions) {

                if (Array.isArray(suggestions)) {
                    finalSuggestions = suggestions;
                } else if (typeof suggestions === "function") {
                    finalSuggestions = suggestions(language);
                } else {
                    debug.error('Suggestions type not supported');
                    return;
                }

                finalSuggestions = finalSuggestions.map(suggestion => {
                    suggestion.insertText = suggestion.insertText || suggestion.label;
                    return suggestion;
                });

                self._registerDisposable( 'global-suggestions', monaco.languages.registerCompletionItemProvider(language, {
                    provideCompletionItems: function (model, position) {
                        // a completion item provider is unfortunately set to the entire language and is not editor specific
                        // to get around this, whenever this provider is invoked, we check if the editor has focus
                        // if not, it means another editor with this language has focus so we don't return any suggestions
                        if (editor.hasTextFocus()) {
                            word = model.getWordUntilPosition(position);
                            const toReturn = finalSuggestions.map(suggestion => {
                                suggestion.range = {
                                    startLineNumber: position.lineNumber,
                                    endLineNumber: position.lineNumber,
                                    startColumn: word.startColumn,
                                    endColumn: word.endColumn
                                };
                                suggestion.kind = monaco.languages.CompletionItemKind.File;
                                return suggestion;
                            });
                            if(toReturn.length){
                                return {suggestions: toReturn};
                            } else {
                                return;
                            }
                        } else {
                            return;
                        }
                    }
                }));

                // if the language is mle-javascript, we also trigger the suggestions on the dot after apex.env
                if (o.language === LANG_MLE_JAVASCRIPT) {
                    self._registerDisposable( DISP_MLE_ENV, monaco.languages.registerCompletionItemProvider(LANG_JAVASCRIPT, {
                        triggerCharacters: ['.'],
                        provideCompletionItems: function (model, position, context) {
                            if ( // only if this is the editor in question
                                editor.hasTextFocus() &&
                                // only if "." was pressed
                                context.triggerKind === monaco.languages.CompletionTriggerKind.TriggerCharacter &&
                                context.triggerCharacter === '.' &&
                                // only if the dot follows apex.env
                                model.getLineContent(position.lineNumber).substr(0, position.column - 2).endsWith('apex.env')
                            ) {
                                word = model.getWordUntilPosition(position);
                                const toReturn = finalSuggestions.map(suggestion => {
                                    suggestion.range = {
                                        startLineNumber: position.lineNumber,
                                        endLineNumber: position.lineNumber,
                                        startColumn: word.startColumn,
                                        endColumn: word.endColumn
                                    };
                                    suggestion.kind = monaco.languages.CompletionItemKind.File;
                                    return suggestion;
                                });
                                if(toReturn.length){
                                    return {
                                        suggestions: toReturn
                                    };
                                } else {
                                    return;
                                }
                            } else {
                                return;
                            }
                        }
                    }));
                }
            }
        },

        // start of minimap logic
        _setupMinimap: function() {
            const self = this,
                o = self.options,
                editor = self._editor;

            if(self._disposableExists( DISP_MINIMAP )){
                self._disposeDisposable( DISP_MINIMAP );
            }

            function showHideMinimap(){
                if (editor && editor.getModel()) {
                    editor.updateOptions({
                        minimap: {
                            enabled: o.minimap === MINIMAP_ON || (o.minimap === MINIMAP_CONDITIONAL && (editor.getModel().getLineCount() >= 100))
                        }
                    });
                }
            }

            if (o.minimap === MINIMAP_CONDITIONAL){
                self._registerDisposable( DISP_MINIMAP, editor.getModel().onDidChangeContent(util.debounce(showHideMinimap, 1000)) );
            }

            showHideMinimap();
        },
        // end of minimap logic

        _getLineFromMessage: function (message) {
            // An error message has the form "...ORA-06550: line xx, column yy: Error Message...""
            var index = message.indexOf('ORA-06550');
            if (index > -1) {
                var str = message.slice(index);
                var parsedError = str.match(/\d{1,}/g);
                var lineNumber = parseInt(parsedError[1], 10);
                var columnNumber = parseInt(parsedError[2], 10);
                if (isNaN(lineNumber)) {
                    lineNumber = 0;
                }
                if (isNaN(columnNumber)) {
                    columnNumber = 0;
                }

                return { lineNumber: lineNumber, columnNumber: columnNumber };
            } else {
                return null;
            }
        },

        // reveals line in center and sets the cursor at the given position
        // line and column must be 1-based
        // pass focus = false, if the editor should not be focused after the position was set
        //      this can avoid console errors in cases where the focus is trapped somewhere else,
        //      like alert dialogs. see bug #33264302
        gotoPosition: function (line, column, focus = true) {
            var self = this,
                editor = self._editor;

            if (editor) {
                editor.revealLineInCenter(line);
                editor.setPosition({
                    lineNumber: line,
                    column: column
                });
                if( focus ){
                    setTimeout(function () {
                        editor.focus();
                    }, 100);
                }
            }
        },

        _updateNotifications: function () {
            var self = this,
                options = self.options,
                list$ = $('<ul/>');

            if (options.errors.length || options.warnings.length) {
                for (let i = 0; i < options.errors.length; i++) {
                    let message = util.escapeHTML(options.errors[i]),
                        lineObject = self._getLineFromMessage(message),
                        listItem$ = $('<li class="is-error" style="cursor: pointer;"></li>');

                    if (lineObject) {
                        listItem$.append($(`<a data-line="${lineObject.lineNumber}" data-column="${lineObject.columnNumber}"></a>`).html(message));
                    } else {
                        listItem$.html(message);
                    }

                    list$.append(listItem$);
                }
                for (let i = 0; i < options.warnings.length; i++) {
                    let message = util.escapeHTML(options.warnings[i]),
                        lineObject = self._getLineFromMessage(message),
                        listItem$ = $('<li class="is-warning" style="cursor: pointer;"></li>');

                    if (lineObject) {
                        listItem$.append($(`<a data-line="${lineObject.lineNumber}" data-column="${lineObject.columnNumber}"></a>`).html(message));
                    } else {
                        listItem$.html(message);
                    }

                    list$.append(listItem$);
                }

                self.showNotification(list$);
            } else {
                self.resetNotification();
            }

            $('a[data-line]', list$).on('click', function () {
                self.gotoPosition($(this).data('line'), $(this).data('column'));
            });

            self._updateLineMessages();
        },

        _updateLineMessages: function () {
            var self = this,
                options = self.options,
                editor = self._editor,
                lineMessages = [],
                lineMessage;

            // error have format:     ORA-20999: Failed to parse SQL query! <p>ORA-06550: line 6, column 5: ORA-00942: table or view does not exist</p>
            // warnings have format:  ORA-06550: line 1, column 78: PL/SQL: ORA-00942: table or view does not exist
            function cleanMessageForInline(options) {
                var newMessage = options.message.split(':').slice(options.afterColonIndex).join(':').trim();

                if (options.removeHtmlTags) {
                    newMessage = util.stripHTML(newMessage);
                }

                if (newMessage.length) {
                    return newMessage;
                } else {
                    return options.message;
                }
            }

            for (let i = 0; i < options.errors.length; i++) {
                lineMessage = self._getLineFromMessage(options.errors[i]);
                if (lineMessage) {
                    lineMessages.push({
                        startLineNumber: lineMessage.lineNumber,
                        endLineNumber: lineMessage.lineNumber,
                        startColumn: lineMessage.columnNumber,
                        endColumn: 1000,
                        message: cleanMessageForInline({
                            message: options.errors[i],
                            afterColonIndex: 4,
                            removeHtmlTags: true
                        }),
                        severity: monaco.MarkerSeverity.Error
                    });
                }
            }
            for (let i = 0; i < options.warnings.length; i++) {
                lineMessage = self._getLineFromMessage(options.warnings[i]);
                if (lineMessage) {
                    lineMessages.push({
                        startLineNumber: lineMessage.lineNumber,
                        endLineNumber: lineMessage.lineNumber,
                        startColumn: lineMessage.columnNumber,
                        endColumn: 1000,
                        message: cleanMessageForInline({
                            message: options.warnings[i],
                            afterColonIndex: 3,
                            removeHtmlTags: false
                        }),
                        severity: monaco.MarkerSeverity.Warning
                    });
                }
            }

            monaco.editor.setModelMarkers(editor.getModel(), 'lineMessages', lineMessages);

            if (lineMessages.length) {
                self.gotoPosition(lineMessages[0].startLineNumber, lineMessages[0].startColumn);
            }
        },

        _initNotificationBar: function () {
            var self = this,
                out = util.htmlBuilder(),
                closeId = this.baseId + '_mClose';

            out.markup(`<div class="${C_NOTIFICATION}" style="display:none;"><div class="${C_NOTIFICATION_MSG}"></div>`);

            out.markup("<button")
                .attr("id", closeId)
                .attr("title", MSG_CLOSE)
                .attr("aria-label", MSG_CLOSE)
                .optionalBoolAttr("disabled", false)
                .attr("class", "a-Button a-Button--noLabel a-Button--withIcon a-Button--small a-CodeEditor-searchBar-closeButton")
                .markup(" type='button'>")
                .markup("<span class='a-Icon ")
                .attr('ui-icon-closethick')
                .markup("' aria-hidden='true'></span></button>");

            out.markup('</div>');
            this.element.find(SEL_TOOLBAR).after(out.toString());
            $('#' + closeId).click(function () {
                $(this).parent().hide();
                self.resize();
            });
        },

        _queryBuilder: function () {
            var fn = this.options.queryBuilder;
            if (fn) {
                fn(this, this.getValue());
            }
        },

        _codeComplete: function () {

            // ajax-based code complete is only allowed for sql and mle-javascript, and only if a callback has been provided
            if (!(this.options.codeComplete && [LANG_SQL, LANG_MLE_JAVASCRIPT].includes(this.options.language))) {
                return;
            }

            var self = this,
                language = self.options.language,
                editor = self._editor,
                model = editor.getModel(),
                currentPosition = editor.getPosition(),
                lineValue = model.getLineContent(currentPosition.lineNumber),
                word, parts, search, parent, grantParent,
                isItem = false,
                $spinner,
                elem = self.element[0];

            // the builtin model.getWordAtPosition and getWordAroundPosition functions
            // to not return values such as "x.y.z", only "z"
            // so we use our own function
            function getWordAround(s, pos) {
                // make pos point to a character of the word
                while (s[pos] === ' '){
                    pos = pos - 1;
                }
                // find the space before that word
                // (add 1 to be at the begining of that word)
                // (note that it works even if there is no space before that word)
                pos = s.lastIndexOf(' ', pos) + 1;
                // find the end of the word
                var end = s.indexOf(' ', pos);
                if (end === -1) {
                    end = s.length; // set to length if it was the last word
                }
                // return the result
                return s.substring(pos, end);
            }

            word = getWordAround(lineValue, currentPosition.column - 1);
            parts = word.split('.');

            if (language === LANG_SQL) {
                if (parts.length === 1 && [':', '&'].includes(word.charAt(0))) {
                    isItem = true;
                    search = word.slice(1);
                } else {
                    parent = parts[parts.length - 2];
                    grantParent = parts[parts.length - 3];
                }
            } else if (language === LANG_MLE_JAVASCRIPT) {
                if (parts.length === 3 && parts[0] === 'apex' && parts[1] === 'env') {
                    isItem = true;
                    search = parts[2];
                    parent = undefined;
                    grantParent = undefined;
                } else {
                    self._triggerMonacoAction('editor.action.triggerSuggest');
                    return;
                }
            }

            // function invoked when the autocomplete items are returned
            // pData has to be in the format:
            //   [
            //     type:      "string", (template, application_item, page_item, package, procedure, function, constant, variable, type, table, view)
            //     title:     "string",
            //     className: "string",
            //     completions: [
            //       { d: "string", r: "string" } or "string"
            //     ]
            //   ]
            function _success(pData) {

                apex.util.delayLinger.finish('autocompleteSpinner', function () {
                    if ($spinner) {
                        $spinner.remove();
                        $spinner = null;
                    }
                });

                var kinds = monaco.languages.CompletionItemKind;

                // TODO: find more fitting icons
                var monacoTypes = {
                    template: kinds.File,
                    application_item: kinds.File,
                    page_item: kinds.File,
                    package: kinds.File,
                    procedure: kinds.Method,
                    function: kinds.Function,
                    constant: kinds.File,
                    variable: kinds.File,
                    type: kinds.File,
                    table: kinds.File,
                    view: kinds.File,
                    keyword: kinds.File,
                    sequence: kinds.File
                };

                var type,
                    completion,
                    completions = [];
                for (var i = 0; i < pData.length; i++) {
                    type = pData[i];

                    for (var j = 0; j < type.completions.length; j++) {
                        completion = type.completions[j];
                        var text = completion.r || completion;

                        completions.push({
                            label: (completion.d || completion),
                            insertText: text,
                            detail: type.title,
                            kind: monacoTypes[type.type]
                        });
                    }
                }

                // alreadyShown ensures temporary completions are only used by the completionItemProvider once: now.
                // it would be more elegant to dispose the registerCompletionItemProvider disposable when we're done with the autocomplete
                // but we don't have enough info on this, risking to hide the popup right after it was shown
                var alreadyShown = false;
                self._registerDisposable( 'ajax-suggestions', monaco.languages.registerCompletionItemProvider(model._languageIdentifier.language, {
                    provideCompletionItems: function () {
                        if (self.element.hasClass(C_ACTIVE) && !alreadyShown) {
                            alreadyShown = true;
                            if(completions.length){
                                return {
                                    suggestions: completions
                                };
                            } else {
                                return;
                            }
                        } else {
                            return;
                        }
                    }
                }));

                // show the completion suggestions menu by force
                self._triggerMonacoAction('editor.action.triggerSuggest');
            } // _success


            util.delayLinger.start('autocompleteSpinner', function () {
                $spinner = util.showSpinner(elem);
            });

            self.options.codeComplete({
                type: isItem ? 'item' : '',
                search: isItem ? search : parts[parts.length - 1],
                parent: parent,
                grantParent: grantParent
            }, _success);
        },

        _validateCode: function () {
            if (!this.options.validateCode) {
                return;
            }

            var self = this;

            self.options.validateCode(self.getValue(), function (results) {
                results = $.extend({}, { errors: [], warnings: [] }, results);
                self._setOption('errors', results.errors);
                self._setOption('warnings', results.warnings);
                if (results.errors.length === 0 && results.warnings.length === 0) {
                    // indicate that all is well
                    self.showSuccessNotification(util.escapeHTML(MSG_VALIDATION_SUCCESS));
                }
            });
        },

        _createContext: function () {
            var context = actions.createContext('codeEditor', this.element[0]);
            return context;
        },

        _notifyPreferenceChange: function () {
            var self = this,
                element = self.element[0];
            self._trigger('preferencesChanged', $.Event('click', { target: element }));
        },

        _populateContext: function () {
            var self = this,
                o = self.options,
                language = o.language,
                editor = self._editor,
                context = self._context;

            function updateIndentation() {
                editor.getModel().updateOptions({
                    insertSpaces: o.tabsInsertSpaces,
                    tabSize: o.tabSize,
                    indentSize: o.indentSize
                });
            }

            context.add([{
                name: 'undo',
                action: function () {
                    self._triggerMonacoAction('undo');
                }
            }, {
                name: 'redo',
                action: function () {
                    self._triggerMonacoAction('redo');
                }
            }, {
                name: 'find',
                action: function () {
                    // findReplace does not exist in readOnly mode, so we must invoke the regular find
                    self._triggerMonacoAction(o.readOnly ? 'actions.find' : 'editor.action.startFindReplaceAction');
                }
            }, {
                name: 'theme',
                get: function () {
                    return o.theme;
                },
                set: function (v) {
                    o.theme = v;
                    setTheme( v );
                    self._notifyPreferenceChange();
                },
                choices: OPTIONS_THEME
            }, {
                name: 'minimap',
                get: function () {
                    return o.minimap;
                },
                set: function (v) {
                    o.minimap = v;
                    self._setupMinimap();
                    self._notifyPreferenceChange();
                },
                choices: OPTIONS_MINIMAP
            }, {
                name: 'tabs-insert-spaces',
                label: MSG_INDENT_WITH_TABS,
                get: function () {
                    return o.tabsInsertSpaces;
                },
                set: function (v) {
                    o.tabsInsertSpaces = v;
                    updateIndentation();
                    self._notifyPreferenceChange();
                }
            }, {
                name: 'tab-size',
                get: function () {
                    return o.tabSize;
                },
                set: function (v) {
                    o.tabSize = v;
                    updateIndentation();
                    self._notifyPreferenceChange();
                },
                choices: OPTIONS_TAB_SIZE
            }, {
                name: 'indent-size',
                get: function () {
                    return o.indentSize;
                },
                set: function (v) {
                    o.indentSize = v;
                    updateIndentation();
                    self._notifyPreferenceChange();
                },
                choices: OPTIONS_INDENT_SIZE
            }, {
                name: 'show-ruler',
                label: MSG_SHOW_RULER,
                get: function () {
                    return o.ruler;
                },
                set: function (v) {
                    self._setOption('ruler', v);
                    self._notifyPreferenceChange();
                }
            }, {
                name: 'line-numbers',
                label: MSG_SHOW_LINE_NUMBERS,
                get: function () {
                    return o.lineNumbers;
                },
                set: function (v) {
                    self._setOption('lineNumbers', v);
                    self._notifyPreferenceChange();
                }
            }, {
                name: 'whitespace',
                label: MSG_WHITESPACE,
                get: function () {
                    return o.whitespace;
                },
                set: function (v) {
                    self._setOption('whitespace', v);
                    self._notifyPreferenceChange();
                }
            }, {
                name: 'show-suggestions',
                label: MSG_SUGGESTIONS,
                get: function () {
                    return o.showSuggestions;
                },
                set: function (v) {
                    self._setOption('showSuggestions', v);
                    self._notifyPreferenceChange();
                }
            }, {
                name: 'accessibility-mode',
                label: MSG_ACCESSIBILITY_MODE,
                get: function () {
                    return o.accessibilityMode;
                },
                set: function (v) {
                    self._setOption('accessibilityMode', v);
                    self._notifyPreferenceChange();
                }
            }]);

            if (o.queryBuilder) {
                context.add({
                    name: 'query-builder',
                    hide: language !== LANG_SQL,    // initial hidden state
                    action: function () {
                        self._queryBuilder();
                    }
                });
            }

            if (o.codeComplete) {
                context.add({
                    name: 'code-complete',
                    hide: ![LANG_SQL, LANG_MLE_JAVASCRIPT].includes(language),    // initial hidden state
                    action: function () {
                        self._codeComplete();
                    }
                });

                editor.addAction({
                    id: 'apex-code-complete',
                    label: MSG_HINT,
                    keybindings: [getUniversalCtrlKey() | monaco.KeyCode.Space],
                    precondition: 'dbHint',
                    run: function () {
                        self._codeComplete();
                    }
                });
            }

            if (o.validateCode) {
                context.add({
                    name: 'validate',
                    hide: language !== LANG_SQL,    // initial hidden state
                    action: function () {
                        self._validateCode();
                    }
                });
                editor.addAction({
                    id: 'apex-code-validate',
                    label: MSG_VALIDATE,
                    keybindings: [getUniversalCtrlKey() | monaco.KeyMod.Alt | monaco.KeyCode.KEY_V],
                    precondition: 'isSql',
                    run: function () {
                        self._validateCode();
                    }
                });
            }

            // logic for enabling or disabling the undo/redo buttons
            (function () {
                // save states for undo/redo button disabling logic
                var initialVersion = editor.getModel().getAlternativeVersionId(),
                    currentVersion = initialVersion,
                    lastVersion = initialVersion;

                editor.onDidChangeModelContent(() => {
                    var versionId = editor.getModel().getAlternativeVersionId();
                    // undoing
                    if (versionId < currentVersion) {
                        context.enable('redo');
                        // no more undo possible
                        if (versionId === initialVersion) {
                            context.disable('undo');
                        }
                    } else {
                        // redoing
                        if (versionId <= lastVersion) {
                            // redoing the last change
                            if (versionId === lastVersion) {
                                context.disable('redo');
                            }
                        } else { // adding new change, disable redo when adding new changes
                            context.disable('redo');
                            if (currentVersion > lastVersion) {
                                lastVersion = currentVersion;
                            }
                        }
                        context.enable('undo');
                    }
                    currentVersion = versionId;
                });

                // on itinialization, disable both undo and redo
                context.disable('undo');
                context.disable('redo');
            })();
        },

        _initToolbar: function (container$) {
            var o = this.options,
                helper;

            var config = {
                actionsContext: this._context,
                simple: true,
                data: []
            };

            var undoControls = [];

            undoControls.push({
                type: 'BUTTON',
                title: MSG_UNDO,
                label: MSG_UNDO,
                iconOnly: true,
                icon: 'icon-undo',
                action: 'undo'
            });

            undoControls.push({
                type: 'BUTTON',
                title: MSG_REDO,
                label: MSG_REDO,
                iconOnly: true,
                icon: 'icon-redo',
                action: 'redo'
            });

            config.data.push({
                id: 'undoControls',
                align: 'start',
                groupTogether: true,
                controls: undoControls
            });

            var searchControls = [];
            searchControls.push({
                type: 'BUTTON',
                title: MSG_FIND,
                label: MSG_FIND,
                iconOnly: true,
                icon: 'icon-cm-find',
                action: 'find'
            });

            config.data.push({
                id: 'searchControls',
                align: 'start',
                groupTogether: true,
                controls: searchControls
            });

            if (o.queryBuilder || o.codeComplete || o.validateCode) {
                var helperGroup = {
                    id: 'helperControls',
                    align: 'start',
                    groupTogether: true,
                    controls: []
                };

                if (o.queryBuilder) {
                    helperGroup.controls.push({
                        type: 'BUTTON',
                        title: MSG_QUERY_BUILDER,
                        label: MSG_QUERY_BUILDER,
                        iconOnly: true,
                        icon: 'icon-cm-query-builder',
                        action: 'query-builder'
                    });
                }

                if (o.codeComplete) {
                    helper = MSG_HINT + ' - Ctrl+Space';
                    helperGroup.controls.push({
                        type: 'BUTTON',
                        title: helper,
                        label: helper,
                        iconOnly: true,
                        icon: 'icon-cm-autocomplete',
                        action: 'code-complete'
                    });
                }

                if (o.validateCode) {
                    helper = MSG_VALIDATE + ' - Ctrl+Alt+V';
                    helperGroup.controls.push({
                        type: 'BUTTON',
                        title: helper,
                        label: helper,
                        iconOnly: true,
                        icon: 'icon-cm-validate',
                        action: 'validate'
                    });
                }

                config.data.push(helperGroup);
            }

            config.data.push({
                id: 'menuControls',
                align: 'end',
                groupTogether: false,
                controls: [{
                    type: 'MENU',
                    title: MSG_SETTINGS,
                    label: MSG_SETTINGS,
                    iconOnly: true,
                    icon: 'icon-gear',
                    menu: {
                        items: [{
                            type: 'subMenu',
                            label: MSG_INDENTATION,
                            menu: {
                                items: [{
                                    type: 'toggle',
                                    action: 'tabs-insert-spaces'
                                }, {
                                    type: 'subMenu',
                                    label: MSG_TAB_SIZE,
                                    menu: {
                                        items: [{
                                            type: 'radioGroup',
                                            action: 'tab-size'
                                        }]
                                    }
                                }, {
                                    type: 'subMenu',
                                    label: MSG_INDENT_SIZE,
                                    menu: {
                                        items: [{
                                            type: 'radioGroup',
                                            action: 'indent-size'
                                        }]
                                    }
                                }
                                ]
                            }
                        }, {
                            type: 'subMenu',
                            label: MSG_THEMES,
                            menu: {
                                items: [{
                                    type: 'radioGroup',
                                    action: 'theme'
                                }]
                            }
                        }, {
                            type: 'subMenu',
                            label: MSG_MINIMAP,
                            menu: {
                                items: [{
                                    type: 'radioGroup',
                                    action: 'minimap'
                                }]
                            }
                        }, {
                            type: 'toggle',
                            action: 'accessibility-mode'
                        }, {
                            type: 'toggle',
                            action: 'line-numbers'
                        }, {
                            type: 'toggle',
                            action: 'show-ruler'
                        }, {
                            type: 'toggle',
                            action: 'show-suggestions'
                        }, {
                            type: 'toggle',
                            action: 'whitespace'
                        }
                    ]
                    }
                }]
            });

            container$.toolbar(config);
        },

        // should be called on widget initialization and whenever the language changes
        // the language should not be sanatized. ie. we expect the mle- prefix in the case of mle-javascript
        _evaluateContextKeys: function (language) {

            var keys = this._contextKeys,
                editor = this._editor;

            if (!keys.isSql) {
                // dummy assignment
                keys.isSql = editor.createContextKey('isSql', true);
            }
            if (!keys.dbHint) {
                // dummy assignment
                keys.dbHint = editor.createContextKey('dbHint', true);
            }

            keys.isSql.set(language === LANG_SQL);
            keys.dbHint.set([LANG_SQL, LANG_MLE_JAVASCRIPT].includes(language));
        },

        _setLanguage: function (language) {
            var self = this,
                editor = self._editor,
                o = self.options,
                context = self._context;

            // remove any error messages
            monaco.editor.setModelMarkers(editor.getModel(), 'lineMessages', []);
            self.resetNotification();

            // reconfigure editor specific suggestions
            self._configureCustomSuggestionProvider();

            function hide(actionName) {
                if (context.lookup(actionName)) {
                    context.hide(actionName);
                }
            }
            function show(actionName) {
                if (context.lookup(actionName)) {
                    context.show(actionName);
                }
            }

            if ([LANG_SQL, LANG_MLE_JAVASCRIPT].includes(language)) {
                show('code-complete');
            } else {
                hide('code-complete');
            }

            if (language === LANG_SQL) {
                show('validate');
                show('query-builder');
            } else {
                hide('validate');
                hide('query-builder');
            }

            self._evaluateContextKeys(language);

            loadLanguageDependencies(language, o.pageInfo);

            monaco.editor.setModelLanguage(editor.getModel(), language.replace('mle-', ''));
        },

        /*
         * Add warnings to page item references which don't exist
         * This only takes into account page items of the current page and the global page
         * as these are the only ones we currently know about
         * We assume the conventional Pnnn_[alphanum_]* name types
         */
        _initializeItemValidator: function () {
            var self = this,
                editor = self._editor,
                model = editor.getModel(),
                o = self.options,
                pageId = o.pageInfo.pageId,
                itemRegex = new RegExp('P(' + pageId + '|0)_[A-Z0-9_]*', 'gi'),
                warnings = [];

            var pageItems = o.pageInfo.items.map( item => item.name.toLowerCase() );

            var verify = util.debounce(function () {
                warnings = [];
                // always check if the editor still exists as we work with debounced functions
                // which might fire after the editor was closed
                if (editor && model) {
                    // loop through each line
                    var lineCount = model.getLineCount();

                    for (var lineNumber = 1; lineNumber <= lineCount; lineNumber++) {
                        var lineContent = model.getLineContent(lineNumber);

                        var matches = lineContent.matchAll(itemRegex);

                        for (var match of matches) {
                            var foundItem = match[0].toLowerCase(),
                                itemFoundAt = match.index + 1; // adjusting for 1-based index

                            if (!pageItems.includes(foundItem)) {
                                warnings.push({
                                    startLineNumber: lineNumber,
                                    endLineNumber: lineNumber,
                                    startColumn: itemFoundAt,
                                    endColumn: itemFoundAt + foundItem.length,
                                    message: lang.formatMessage('CODE_EDITOR.ITEM_DOES_NOT_EXIST', foundItem),
                                    severity: monaco.MarkerSeverity.Warning
                                });
                            }
                        }
                    }
                    monaco.editor.setModelMarkers(model, 'itemValidations', warnings);
                }
            }, 100);

            // validating whenever the editor changed content, but debounced at 100 ms
            model.onDidChangeContent(verify);

            // valiadating on editor load as well
            verify();
        },

        /*
         * Configuration of various shortcuts and extra editor actions
         */
        _initializeCustomActions: function () {
            var self = this,
                editor = self._editor,
                universalCtrlKey = getUniversalCtrlKey();

            // adds a keybinding to a specific action
            // see https://github.com/microsoft/monaco-editor/issues/102 for more
            // there is no official API to do this, so we add another action with the same id and info
            // the new action+keybinding will appear in the Command Palette
            // but if the old action had a keybinding itself, it will still work
            function overrideActionKeybinding(actionId, newKeyBinding) {
                var action = editor.getAction(actionId);
                editor.addAction({
                    id: actionId,
                    label: action.label,
                    alias: action.alias,
                    keybindings: [newKeyBinding],
                    run: function () {
                        action.run();
                    }
                });
            }

            // overriding the keybindings of the Transform to Upper/Lower actions
            overrideActionKeybinding('editor.action.transformToUppercase', universalCtrlKey | monaco.KeyMod.Alt | monaco.KeyCode.KEY_U);
            overrideActionKeybinding('editor.action.transformToLowercase', universalCtrlKey | monaco.KeyMod.Alt | monaco.KeyCode.KEY_L);

            editor.addAction({
                id: 'editor.action.toggleSmartCase',
                label: MSG_TRANSFORM_CODE_CASE,
                alias: MSG_TRANSFORM_CODE_CASE,
                keybindings: [universalCtrlKey | monaco.KeyMod.Alt | monaco.KeyCode.KEY_C],
                run: function () {
                    // adapted from linesOperators.ts AbstractCaseAction
                    var selections = editor.getSelections();
                    if (selections === null) {
                        return;
                    }
                    var model = editor.getModel();
                    if (model === null) {
                        return;
                    }

                    var edits = [];

                    for (var i = 0; i < selections.length; i++) {
                        var selection = selections[i];
                        if (!selection.isEmpty()) {
                            var text = model.getValueInRange(selection);
                            var transformedText = transformCodeCase(text);
                            edits.push({
                                range: selection,
                                text: transformedText
                            });
                        }
                    }

                    editor.pushUndoStop();
                    editor.executeEdits(null, edits);
                    editor.pushUndoStop();
                }
            });

            // By default, the Command Palette is opened by hitting the F1 key
            // Also adding the CTRL/CMD+Shift+P shortcut to match VS Code
            // Intentionally not using universalCtrlKey here
            editor.addCommand(monaco.KeyMod.CtrlCmd | monaco.KeyMod.Shift | monaco.KeyCode.KEY_P, () => {
                self._triggerMonacoAction('editor.action.quickCommand');
            });
        }
    });

    $.apex.codeEditor.preferencesObjectFromString = function( optionsString ) {
        var rawOptions,
            finalOptions = {};
        
        if(!optionsString){
            return {};
        }

        try {
            rawOptions = JSON.parse(optionsString);
        } catch (e) {
            debug.warn('Code Editor: could not parse optionsString');
            return {};
        }

        $.each(PREFERENCE_KEYS, function (i, item) {
            var value = rawOptions[item],
                key = item;

            if (key === 'theme') {
                if (!OPTIONS_THEME.map(option => option.value).includes(value)) {
                    debug.warn('Code Editor: Bad theme ignored: ' + value);
                    return;
                }
            } else if (['indentSize', 'tabSize'].includes(key)) {
                if (!['2', '3', '4', '8'].includes(value)) {
                    debug.warn('Code Editor: Bad number ignored: ' + value);
                    return;
                }
            } else if (['tabsInsertSpaces', 'ruler', 'accessibilityMode', 'lineNumbers', 'whitespace', 'showSuggestions'].includes(key)){
                value = !!value;
            } else if (key === 'minimap'){
                if (!OPTIONS_MINIMAP.map(option => option.value).includes(value)) {
                    debug.warn('Code Editor: Bad minimap ignored: ' + value);
                    return;
                }
            } else {
                debug.warn('Code Editor: Unknown preference: ' + key);
                return;
            }

            finalOptions[key] = value;
        });

        return finalOptions;
    };

})(apex.jQuery, apex.util, apex.lang, apex.locale, apex.debug, apex.actions, apex.server, apex.env);