/*!
 Copyright (c) 2013, 2021, Oracle and/or its affiliates. All rights reserved.
*/
/**
 * The {@link apex.widget.ckeditor5} is used for the Rich Text Editor widget of Oracle APEX.
 * Internally the CKEditor http://www.ckeditor.com is used.
 * See the CKEditor documentation for available options.
 **/
/* global ClassicEditor */
(function (item, lang, util, widget, debug, $) {
    "use strict";
    /**
     * @param {String}   pSelector                  jQuery selector to identify APEX page item(s) for this widget.
     * @param {Object}   [pOptions]                 Optional object holding overriding options.
     * @param {Function} [pPluginInitJavascript]    Optional function which allows overriding or extending of the widget options.
     *
     * @function ckeditor5
     * @memberOf apex.widget
     * */
    widget.ckeditor5 = function (pSelector, pOptions, pPluginInitJavascript) {

        // Based on our custom settings, add additional properties to the rich text editor options
        var options = $.extend(true, {
            mode: 'html',               // or markdown
            label: null,
            toolbar: 'intermediate',    // or basic or full
            minHeight: 180,
            maxHeight: null,
            // function to be executed right after editor initialization
            // function(editor){
            //     // perform any actions on the editor before any page load actions
            // }
            executeOnInitialization: null,
            // wraps the content in <div class="ck-content"></div>
            // if it contains any classes or content needing styling
            automaticStyleWrap: true,
            // what to return as Display Value, e.g in IG
            // clean: removes all html tags
            // escape: escapes the value
            // unchanged: returns the value as is
            displayValueMode: 'clean',
            editorOptions: {
                language: 'en',
                extraPlugins: [],
                removePlugins: [],
                toolbar: [],
                codeBlock: {
                    languages: [
                        { language: 'plaintext', label: 'Plaintext' },
                        { language: 'html', label: 'HTML' },
                        { language: 'css', label: 'CSS' },
                        { language: 'js', label: 'JavaScript' },
                        { language: 'sql', label: 'SQL' }
                    ]
                },
                table: {
                    contentToolbar: [
                        'tableColumn',
                        'tableRow',
                        'mergeTableCells',
                        'tableCellProperties',
                        'tableProperties'
                    ]
                },
                fontSize: {
                    options: [
                        'tiny',
                        'small',
                        'default',
                        'big',
                        'huge'
                    ],
                    supportAllValues: false
                }
            }
        }, pOptions);

        if (options.mode === 'html') {
            if (options.toolbar === 'basic') {
                options.editorOptions.toolbar = [
                    'bold', 'italic', '|',
                    'bulletedList', 'numberedList', '|',
                    'undo', 'redo'
                ];
            } else if (options.toolbar === 'intermediate') {
                options.editorOptions.toolbar = [
                    'heading', '|',
                    'bold', 'italic', 'underline', 'strikethrough', '|',
                    'link', 'bulletedList', 'numberedList', '|',
                    'blockQuote', 'insertTable', '|',
                    'undo', 'redo'
                ];
            } else if (options.toolbar === 'full') {
                options.editorOptions.extraPlugins.push(ClassicEditor.extraPlugins.ListStyle);
                options.editorOptions.toolbar = [
                    'heading', '|',
                    'fontSize', 'fontFamily', 'fontColor', 'fontBackgroundColor', 'highlight', '|',
                    'bold', 'italic', 'underline', 'strikethrough', 'subScript', 'superScript', 'code', '|',
                    'alignment', 'indent', 'outdent', '|',
                    'bulletedList', 'numberedList', 'todoList', '|',
                    'specialCharacters', 'link', 'blockQuote', 'pageBreak', 'horizontalLine', 'insertTable', 'codeBlock', '|',
                    'removeFormat', 'undo', 'redo'
                ];
            }

            options.editorOptions.image = {
                styles: [
                    'full', 'side', 'alignLeft', 'alignCenter', 'alignRight'
                ],
                toolbar: [
                    'imageStyle:alignLeft', 'imageStyle:alignCenter', 'imageStyle:alignRight',
                    '|',
                    'imageResize',
                    '|',
                    'imageTextAlternative'
                ]
            };

        } else if (options.mode === 'markdown') {
            // simply adding the plug-in will enable Markdown mode
            options.editorOptions.extraPlugins.push(ClassicEditor.extraPlugins.Markdown);
            // remove any plug-ins which don't fit Markdown mode
            options.editorOptions.removePlugins.push('ImageToolbar', 'ImageCaption');
            // custom toolbar with only the features Markdown can handle
            options.editorOptions.toolbar = [
                'heading', '|',
                'bold', 'italic', 'code', '|',
                'link', 'bulletedList', 'numberedList', '|',
                'blockQuote', 'codeBlock', '|',
                'undo', 'redo'
            ];
        }

        // JavaScript Initialization Code
        if ($.isFunction(pPluginInitJavascript)) {
            var changeOptions = pPluginInitJavascript(options);
            if (changeOptions) {
                options = changeOptions;
            }
        }

        // Instantiate the CKEditor
        $(pSelector).each(function () {
            var textArea$ = $(this),
                widgetId = this.id + '_WIDGET',
                itemName = this.id,
                initialData = textArea$.val(),
                editor,
                setFocus,
                deferredObject;

            textArea$
                .hide()
                .parent().append(`<div id="${util.escapeHTMLAttr(widgetId)}"></div>`);

            setFocus = function () {
                editor.focus();
            };

            deferredObject = item.create(itemName, {
                delayLoading: true,
                enable: function () {
                    editor.isReadOnly = false;
                },
                disable: function () {
                    editor.isReadOnly = true;
                },
                setValue: function (pValue) {
                    // using == on purpose to catch undefined
                    editor.setData(pValue == null ? '' : '' + pValue);
                },
                getValue: function () {
                    var data = editor.getData(),
                        stylePrefix = '<div class="ck-content">',
                        stylePostfix = '</div>';

                    // For html
                    // if the content contains any classes (we assume all classes are used for styling)
                    // and only for elements blockquote, code, hr and pre (as of v25.0.0)
                    // We wrap the content in <div class="ck-content"></div>
                    // so the content will keep its formatting outside of the editor as well
                    if (options.mode === 'html' &&
                        options.automaticStyleWrap &&
                        /( class="|<blockquote|<code|<hr|<pre)/.test(data) &&
                        !data.startsWith(stylePrefix)
                    ) {
                        data = stylePrefix + data + stylePostfix;
                    }

                    return data;
                },
                displayValueFor: function (value) {
                    if (options.displayValueMode === 'unchanged') {
                        return value;
                    } else if (options.displayValueMode === 'escape') {
                        return util.escapeHTML(value);
                    } else {
                        // clean and default
                        return util.stripHTML(value.replace(/<p>|<br>/g, ' ')).replace(/&nbsp;/g, ' ');
                    }
                },
                setFocusTo: function () {
                    return { focus: setFocus };
                },
                isChanged: function () {
                    return initialData !== editor.getData();
                },
                getEditor: function () {
                    return editor;
                },
                getValidity: function () {
                    // must synchronize with original textarea
                    // to reuse the native validation functionality
                    textArea$.val(this.getValue());
                    return textArea$[0].validity || { valid: true };
                },
                getValidationMessage: function () {
                    var message = '',
                        validMessage = textArea$.attr('data-valid-message');
                    if (!this.getValidity().valid && validMessage) {
                        message = validMessage;
                    } else {
                        // must synchronize with original textarea
                        textArea$.val(this.getValue());
                        message = textArea$[0].validationMessage || '';
                    }
                    return message;
                },
                getPopupSelector: function () {
                    // when editing in the Interactive Grid
                    // the editor would be hidden when various editor popup-s would receive focus
                    // e.g. the Link popup
                    // providing a general selector that all popups share fixes this
                    return '.ck.ck-balloon-panel';
                }
            });

            options.editorOptions.initialData = textArea$.val();

            ClassicEditor
                .create(document.getElementById(widgetId), options.editorOptions)
                .then(function (newEditor) {
                    editor = newEditor;

                    editor.editing.view.change(writer => {
                        writer.setStyle('min-height', (options.minHeight || 150) + 'px', editor.editing.view.document.getRoot());
                        if (options.maxHeight) {
                            writer.setStyle('max-height', options.maxHeight + 'px', editor.editing.view.document.getRoot());
                        }
                    });

                    var snapshot;
                    editor.editing.view.document.on('change:isFocused', function (evt, data, isFocused) {
                        if (isFocused) {
                            snapshot = editor.getData();
                        } else if (editor.getData() !== snapshot) {
                            textArea$.trigger('change');
                        }
                    });

                    // the aria label seems to not be configurable, so we simply override it
                    var newLabel = lang.formatMessage('APEX.RICH_TEXT_EDITOR.ACCESSIBLE_LABEL', options.label || '');
                    $('label.ck.ck-label.ck-voice-label', editor.ui.view.element).text(newLabel);

                    if (options.executeOnInitialization) {
                        options.executeOnInitialization(editor);
                    }

                    deferredObject.resolve();
                })
                .catch(function (error) {
                    debug.error(error);
                });

            // register focus handling, so when the non-displayed textarea of the CKEditor
            // receives focus, focus is moved to the editor.
            textArea$.focus(setFocus);
        });
    }; // ckeditor5

    /*
     * Allow various editor popups to work in jQuery UI dialogs
     */
    if ( $.ui.dialog ) {
        $.widget( "ui.dialog", $.ui.dialog, {
            _allowInteraction: function( event ) {
                return $( event.target ).closest( ".ck" ).length > 0 || this._super( event );
            }
        });
    }

})(apex.item, apex.lang, apex.util, apex.widget, apex.debug, apex.jQuery);