/*!
 Date Picker - APEX widget for picking dates, supports rendering as either JET or Native date form controls
 Copyright (c) 2021, Oracle and/or its affiliates. All rights reserved.
 */
/* eslint-env amd */
/**
 * @fileOverview
 * APEX widget for picking dates, supports rendering as either JET or Native date form controls
 * 
 * Documentation 
 * Note: Both JET and Native HTML date controls require the date to be in ISO Format. This is different to Oracle date format.
 * getValue and setValue both work by using the Oracle date format.
 * 
 * Expected markup
 * 
 * - Display As = Popup / Show Time = No
 *      <oj-input-date
 *          Note: See below for details on relevant attributes for this web component
 *      ></oj-input-date>
 *
 * - Display As = Popup / Show Time = Yes
 *      <oj-input-date-time
 *          Note: See below for details on relevant attributes for this web component
 *      ></oj-input-date-time>
 *
 * - Display As = Inline / Show Time = No
 *      <input type='hidden' id='[item name]_HIDDENVALUE' 
 *          name='[input name from server]' value='[value in Oracle format, used as main value store]'      todo value may not be needed
 *          class='js-ignoreChange' [depends on item ignore change setting]
 *      />
 *      <oj-date-picker
 *          Note: See below for details on relevant attributes for this web component
 *      ></oj-date-picker>  
 *
 * - Display As = Inline / Show Time = Yes:
 *      <input type='hidden' id='[item name]_HIDDENVALUE' 
 *          name='[input name from server]' value='[value in Oracle format, used as main value store]'      todo value may not be needed
 *          class='js-ignoreChange' [depends on item ignore change setting]
 *      />
 *      <oj-date-time-picker
 *          Note: See below for details on relevant attributes for this web component
 *      ></oj-date-time-picker>  
 *
 * - Display As = Native HTML:
 *      <input type=['date' or 'datetime-local': depending on server show time setting] maxlength='10' autocomplete='off' 
 *          Note: In addition see below 'Attributes common to both JET and Native types'
 *      />
 *
 * - Attributes common to both JET and Native types:
 *      - min                               The minimum allowed value, in ISO date format (set by server when min type is static)
 *      - max                               The maximum allowed value, in ISO date format (set by server when max type is static)
 *      - data-dynamic-min                  When min is calculated rather than static, this will be set to an item / DOM ID and 
 *                                          this client-side logic sets up and maintains the min attribute value.
 *      - data-dynamic-max                  When max is calculated rather than static, this will be set to an item / DOM ID and 
 *                                          this client-side logic sets up and maintains the max attribute value.
 *      - value                             Date value in ISO date format
 *
 * - Attributes common to JET types (Display As = Popup or Inline):
 *      - date-picker.number-of-months      JET attribute: Numeric number of months displayed
 *      - date-picker.days-outside-month    JET attribute: Either 'hidden', 'selectable', or 'visible'
 *      - date-picker.week-display          JET attribute: Either 'number' or 'none'
 *      - date-picker.change-month          JET attribute: Either 'select' or 'none'
 *      - date-picker.change-year           JET attribute: Either 'select' or 'none'
 *      - translations.next-text            JET attribute: Message text for next month button
 *      - translations.previous-text        JET attribute: Message text for previous month button
 *      - data-jet-pattern                  Custom attribute that stores the pattern passed to the JET converter object
 *      - data-server-format                Server format used by client-side messaging (todo may change)
 *
 * - Attributes common to JET popup types (Display As = Popup):
 *      - data-size                         Custom attribute used to set JET's shadow DOM input 'size' attribute
 *      - data-maxlength                    Custom attribute used to set JET's shadow DOM input 'maxlength' attribute
 *      - data-name                         Custom attribute used to set JET's shadow DOM input 'name' attribute. 
 *      - Note: For labelling the shadow DOM input with popup types, we add aria-labelledby to the shadow DOM input element. 
 *              See comments in relevant code section below.
 *
 * - Attributes common to JET time types (Display As = Popup or Inline, with Show Time = Yes):
 *      - time-picker.time-increment        Time increment in format hh:mm:ss:SS
 *
 * Note: In APEX we turn off all JET messaging / user assistance, so server sets all the following to 'none':
 *  - display-options.converter-hint
 *  - display-options.messages
 *  - display-options.validator-hint
 * 
 * In addition, it's worth noting that for the JET types, it is possible to use any other  attributes that JET support,
 * (for example dayFormatter), it's just that we don't currently expose those in APEX, so they are not mentioned above.
 *
 * 
 * Assumptions
 * 
 * Depends:
 *
 */

( function( item, $, util, lang ) {
    "use strict";

    const A_MIN                 = "min",
          A_MAX                 = "max",
          P_REQUIRED            = "required",
          JET                   = "jet",
          NATIVE                = "native",
          C_APEX_JET_COMPONENT  = "apex-jet-component",
          C_OJ_DATEPICKER_POPUP = "oj-datepicker-popup",
          S_APEX_JET_COMPONENT  = "." + C_APEX_JET_COMPONENT,
          S_OJ_DATEPICKER_POPUP = "." + C_OJ_DATEPICKER_POPUP,
          OJ_DATE_PICKER        = "oj-date-picker",
          OJ_DATE_TIME_PICKER   = "oj-date-time-picker",
          OJ_INPUT_DATE         = "oj-input-date",
          OJ_INPUT_DATE_TIME    = "oj-input-date-time",
          D_VALID_MESSAGE       = "data-valid-message",
          D_VALID_EXAMPLE       = "data-valid-example",
          D_DYNAMIC_MIN         = "data-dynamic-min",
          D_DYNAMIC_MAX         = "data-dynamic-max",
          D_SIZE                = "data-size",
          D_MAXLENGTH           = "data-maxlength",
          D_NAME                = "data-name",
          D_JET_PATTERN         = "data-jet-pattern";


    // Base item prototype, contains attributes common to all date picker types (eg JET-based, and Native HTML)
    let itemPrototype = {
        // Use defaults for show, hide, enable, disable, isDisabled, setStyle
        getNativeValue: function() {

            // this.node always stores the value in ISO date format, as required by both Native HTML and JET date controls
            // For JET components, this.node is the <oj-> element, for Native HTML, this is the date input
            let nodeValue = this.node.value;

            // Append 'Z' to ISO string to ensure date / time is always in UTC
            if ( nodeValue ) {
                if ( nodeValue.length === 10 ) {
                    nodeValue += "T00:00:00Z";
                } else {
                    nodeValue += "Z";
                }

                // This ISO value can be passed directly to new Date(..), and will return a native date object
                return new Date( nodeValue );
            } else {
                return null;
            }

        },
        isChanged: function() {
            if ( this._initialValue === undefined ) {
                return false;
            }
            return this._initialValue !== this.getValue();
        },
        getValidationMessage: function() {
            let validMessage = this.element.attr( D_VALID_MESSAGE ),
                message = "",
                isRequired = this.element.prop( P_REQUIRED ),
                hasMin = this.element.attr( A_MIN ) !== undefined,
                hasMax = this.element.attr( A_MAX ) !== undefined;

            if ( !this.getValidity().valid ) {
                if ( validMessage ) {
                    message = validMessage;
                } else {

                    // We can determine the message to show by looking at what attributes are present and if the value is required
                    // Note: Now automatically substitutes #LABEL# in getValidationMessage logic in item.js
                    if ( isRequired && this.isEmpty() ) {
                        message = lang.formatMessageNoEscape( "APEX.DATEPICKER_JET.REQUIRED" );
                    } else if ( hasMin && hasMax ) {
                        message = lang.formatMessageNoEscape( "APEX.DATEPICKER_JET.VALUE_MUST_BE_BETWEEN", this._format( this.element.attr( A_MIN ) ), this._format( this.element.attr( A_MAX ) ) );
                    } else if ( hasMin ) {
                        message = lang.formatMessageNoEscape( "APEX.DATEPICKER_JET.VALUE_MUST_BE_ON_OR_AFTER", this._format( this.element.attr( A_MIN ) ) );
                    } else if ( hasMax ) {
                        message = lang.formatMessageNoEscape( "APEX.DATEPICKER_JET.VALUE_MUST_BE_ON_OR_BEFORE", this._format( this.element.attr( A_MAX ) ) );
                    } else {
                        // If none of the above, fallback to basic invalid date with example message
                        message = lang.formatMessageNoEscape( "APEX.DATEPICKER_JET.VALUE_INVALID", this.element.attr( D_VALID_EXAMPLE )); 
                    }
                }
            }
            return message;
        },
        reinit: function( pValue ) {
            let dynamicMinId = this.element.attr( D_DYNAMIC_MIN ),
                dynamicMaxId = this.element.attr( D_DYNAMIC_MAX );
            
            this.setValue( pValue, null, true );

            // If the item has dynamic min / max item, we must set the boundary value again
            if ( dynamicMinId ) {
                setBoundaryValue( A_MIN, dynamicMinId, this );
            }
            if ( dynamicMaxId ) {
                setBoundaryValue( A_MAX, dynamicMaxId, this );
            }
        },
        // item-specific code
        _ignorePropertyChange: false
    };

    // JET item prototype extends the base prototype with some of its own..
    let jetItemPrototype = $.extend( {}, itemPrototype, {
        _datePickerType : JET,
        // displayValueFor: returns value in Oracle date format
        displayValueFor: function( value ) {
            return ( this.isReady() ? this._format( value ) : value );
        },
        getPopupSelector: function() {
            return S_OJ_DATEPICKER_POPUP;
        },
        setFocusTo: function() {

            // todo change when we have the tab stop logic in place

            if ( this._isInline ) {
                return $( this.node ).find( ":focusable" ).first();
            } else {
                return this.node;
            }
        },
        // getValue: returns value in Oracle date format
        getValue: function() {
            
            // If JET has loaded we can call _format, otherwise we fallback to the value rendered by the server
            if ( this.isReady() ) {
                return this._format( this.node.value );
            } else {
                return $( "#" + util.escapeCSS( this.id ) + "_PRE_JET" ).val();
            }
        },
        setValue: function( pValue /*, pDisplayValue, pSuppressChange*/ ) {

            // Note: We always set _ignorePropertyChange flag here, because we never want to trigger the change event here (leave that to wrapSetValue in item.js)
            this._ignorePropertyChange = true;
            
            // Changing the DOM node value directly triggers valueChanged (but not change)
            if ( this.isReady() ) {
                this.node.value = this._parse( pValue );    // transform to ISO format for this.node
            } else {
                // if JET isn't ready yet, use the passed value as is
                this.node.value = pValue;
            }
            
            this._ignorePropertyChange = false;
        },
        getValidity: function() {
            
            // JET keeps track of its validity state with a 'valid' property. This can become invalid for example if
            // the user enters an invalid date manually with keyboard. When this happens, if we were to submit 
            // the page, the new value would not be sent (because JET doesn't update the node value when it's 
            // in this invalid state), and the user would not know anything was wrong. So we need to catch this 
            // on the client-side to show an error message using our client-side validation.

            // Call the JET validate method, this is because the value may have become invalid as a result of an external 
            // change (for example an item has a min item value and the min item was changed to later than the item), and 
            // may not yet be invalid (because JET won't have fired its validation logic in such a case). Validating here 
            // ensures whenever validity is checked, the valid state is up to date.

            // Note: Calling validate causes a valueChange event to trigger if the value has not already been changed by the user.
            // We use the valueChange event to trigger 'change', and as we don't want this to trigger a change, set / reset the 
            // ignore flag here (bug #33082351)

            this._ignorePropertyChange = true;
            this.node.validate();
            this._ignorePropertyChange = false;
    
            return { valid : this.node.valid === "valid" };
        },
        delayLoading: true
    });

    // Native item prototype extends base also...
    let nativeItemPrototype = $.extend( {}, itemPrototype, {
        _datePickerType: NATIVE,
        getValue: function() {
            return this.node.value || "";
        },
        setValue: function( pValue /*, pDisplayValue, pSuppressChange*/ ) {

            // Note: We always set _ignorePropertyChange flag here, because we never want to trigger the change event here (leave that to wrapSetValue in item.js)
            this._ignorePropertyChange = true;
            
            // Changing the DOM node value directly triggers valueChanged (but not change)
            this.node.value = pValue;

            this._ignorePropertyChange = false;
        },
        // Native date types displays the date according to the locale of the user's browser, so be consistent with that by 
        // using native date methods toLocaleString (date and time), and toLocaleDateString (date only)
        // Reference: https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input/date
        displayValueFor: function( value ) {
            let displayDate = "";
            if ( value ) {
                if ( this._hasTime ) {
                    displayDate = new Date( value ).toLocaleString();
                } else {
                    displayDate = new Date( value ).toLocaleDateString();
                }
            }
            return displayDate;
        },
        _format: function( value ) {
            return this.displayValueFor( value );
        }
    });

    // Function returns true if item passed has either dynamic min, or dynamic max data attribute defined
    function hasDynamicBoundary ( item ) {
        let item$ = $( "#" + util.escapeCSS( item.id ) );
        return ( item$.attr( D_DYNAMIC_MIN ) || item$.attr( D_DYNAMIC_MAX ) ? true : false );
    }

    // todo Switch to use apex.date.isValidString, when available
    function isValidDateString ( pDateString ) {
        return !isNaN( Date.parse( pDateString ) );
    }

    function setBoundaryValue( pType, pBoundaryItemId, pCurrentItem ) {
        let boundaryItem = item( pBoundaryItemId ),
            currentItem$ = $( "#" + util.escapeCSS( pCurrentItem.id ) ),
            // Use the boundary item's node value, because min/max needs the date in ISO format. Note
            // this will therefore only work currently if the boundary item is also a new date picker.
            boundaryValue = ( boundaryItem.node ? boundaryItem.node.value : null );

        if ( boundaryValue && isValidDateString( boundaryValue ) ) {
            
            // Native types are stricter about the min/max format:
            // - Native > Date requires the format: yyyy-mm-dd (10 chars)
            // - Native > Date / Time requires the format: yyyy-MM-ddThh:mm (16 chars)
            if ( pCurrentItem._datePickerType === NATIVE ) {
                boundaryValue = boundaryValue.substring( 0, pCurrentItem._hasTime ? 16 : 10 );
            }

            // And finally let's update the min/max value with the new boundary value in ISO string format
            currentItem$.attr( pType, boundaryValue );

        } else {

            // If there is no node value, or the value is not a valid date, then remove the min/max attribute
            currentItem$.removeAttr( pType );
        }
    }   // setBoundaryValue

    // Main attach function
    function attachDatePicker( context$ ) {
        let i,
            nativeItemsWithBoundary = [],
            jetItemsWithBoundary = [],
            jetAsyncItems = [],
            jetElements$ = $(   OJ_INPUT_DATE + S_APEX_JET_COMPONENT + "," + 
                                OJ_INPUT_DATE_TIME + S_APEX_JET_COMPONENT + "," + 
                                OJ_DATE_PICKER + S_APEX_JET_COMPONENT + "," + 
                                OJ_DATE_TIME_PICKER + S_APEX_JET_COMPONENT, context$ ),
            nativeElements$ = $( "input[type=date].apex-item-datepicker,input[type=datetime-local].apex-item-datepicker", context$ );

        // Internal function used to setup min / max boundary values and handlers, when min / max is dynamic
        function initBoundaries( pItems ) {
            let i;

            function init( pType, pBoundaryItemId, pCurrentItemId, pCurrentItem ) {

                // set min/max during setup
                setBoundaryValue( pType, pBoundaryItemId, pCurrentItem );
                
                // and setup change handler to update dpItem$ min/max value when data-dynamic-min/max changes
                $( "#" + util.escapeCSS( pBoundaryItemId ) ).on( "change", function() {
                    setBoundaryValue( pType, pBoundaryItemId, pCurrentItem );
                });
            }  // setup

            for ( i = 0; i < pItems.length; i++ ) {
                let currentItem     = pItems[ i ],
                    id              = currentItem.id,
                    currentItem$    = $( "#" + util.escapeCSS( id ) ), 
                    dynamicMinId    = currentItem$.attr( D_DYNAMIC_MIN ), 
                    dynamicMaxId    = currentItem$.attr( D_DYNAMIC_MAX );

                if ( dynamicMinId ) {
                    init( A_MIN, dynamicMinId, id, currentItem );
                }
                if ( dynamicMaxId ) {
                    init( A_MAX, dynamicMaxId, id, currentItem );
                }
            }
        }

        
        /*
         * First the Native HTML display types. 
         * Note: These DON'T need to delay loading
         */
        nativeElements$.each( ( _, el ) => {
            let thisItem,
                id = el.id;
            
            // create the item 
            item.create( id, nativeItemPrototype );
            thisItem = item( id );
            thisItem._hasTime = el.type === "datetime-local";
            // set _initialValue last, as the getValue call needs to know some of the internal variables
            thisItem._initialValue = thisItem.getValue();

            // Store all items that have a dynamic boundary, as these will need to be processed when all items have been created
            if ( hasDynamicBoundary( thisItem ) ) {
                nativeItemsWithBoundary.push( thisItem );
            }
        });

        // setup min / max values / handlers. Note: Must be done after all item objects have been created
        initBoundaries( nativeItemsWithBoundary );
      

        /*
         * Now the JET based display types
         * Note: These DO delay loading
         */

        jetElements$.each( ( _, el ) => {
            let id          = el.id, 
                jetItem$    = $( el ),
                deferred    = item.create( el.id, jetItemPrototype ),   // Important to create item before JET loads
                thisItem    = item( id );

            require( [ "ojs/ojcore", "ojs/ojconverterutils-i18n", "ojs/ojconverter-datetime", "ojs/ojdatetimepicker" ], function( oj ,ConverterUtilsI18n, DateTimeConverter ) {
                
                // make sure the JET component has completely finished creation
                oj.Context.getContext( this ).getBusyContext().whenReady().then( function() {
                    let jetInput$ = jetItem$.find( "input.oj-text-field-input" );
                    
                    // IG requires popup container can recieve focus
                    $( S_OJ_DATEPICKER_POPUP ).prop( "tabindex", -1 );
        
                    /*
                    * We need to make a few modifications to the shadow DOM <input> rendered by JET:
                    * -    Labelling has to be done here because JET renders the <input> which is what needs labelling. Note: We can't
                    *      use our standard label's FOR attribute, because the oj-* element has the ID of the item name already (which is
                    *      what the label's FOR attribute references).
                    * -    Size, maxlength and name also has to be added to the <input>
                    */
                    jetInput$
                        .attr( "aria-labelledby",   el.id + "_LABEL" )
                        .attr( "size",              jetItem$.attr( D_SIZE ) )
                        .attr( "maxlength",         jetItem$.attr( D_MAXLENGTH ) )
                        .attr( "name",              jetItem$.attr( D_NAME ) );
        
                    // Apply JET converter pattern 
                    el.converter =  new DateTimeConverter.IntlDateTimeConverter({
                        pattern: jetItem$.attr( D_JET_PATTERN )
                    });
        
                    // _format function: Pass in the date value in ISO string format, returns the formatted string according 
                    // to the current converter pattern.
                    // try / catch block is needed because if the value is already formatted (eg on page load), then 
                    // the converter will fail, so in that case just return the value as it is.     
                    // todo retest if this try / catch block is needed
                    thisItem._format = function( value ) {
                        try {
                            return el.converter.format( value );
                        } catch (e) {
                            return value;
                        }
                    };

                    // _parse function: Pass in a string in the Oracle date format, returns the date in ISO format
                    // todo test if this try / catch block is needed
                    thisItem._parse = function( value ) {
                        try {
                            return el.converter.parse( value );
                        } catch (e) {
                            return value;
                        }
                    };

                    // Hook up value property change notification to change event if needed
                    jetItem$.on( "valueChanged", function () {
                        
                        // let's ignore this event when value has been set via setValue (because setValue base item handling handles events here)
                        if ( thisItem._ignorePropertyChange ) {
                            return;
                        }
        
                        // wait until the raw value is updated
                        setTimeout( function () {
                            jetItem$.trigger( "change" );
                        }, 0 );

                    } ).change( function( event ) {
        
                        // suppress change event from internal input element (fires when you change the date with the keyboard, so must be suppressed to avoid duplicate events)
                        if ( event.target !== el ) {
                            event.stopImmediatePropagation();
                        }
                    });
                    
                    // Remove the hint div that shows 'Required' if an item is required
                    // todo causes flicker, handle with CSS
                    jetItem$.find( ".oj-required-inline-container" ).hide();

                    thisItem._hasTime = jetItem$.is( OJ_INPUT_DATE_TIME + "," + OJ_DATE_TIME_PICKER );
                    thisItem._isInline = jetItem$.is( OJ_DATE_PICKER + "," + OJ_DATE_TIME_PICKER );

                    deferred.resolve();

                    // set _initialValue last and after resolve, so getValue goes via the isReady() case
                    thisItem._initialValue = thisItem.getValue();

                });  // context ready

            }); // require

            // Store all items that have a dynamic boundary, as these will need to be processed when all items have been created
            if ( thisItem.whenReady && hasDynamicBoundary( thisItem ) ) {
                jetItemsWithBoundary.push( thisItem );
            }

        }); // each ojElement$

        // Make new array containing all the item's whenReady deferred objects, as we need to wait until they are all ready
        for ( i = 0; i < jetItemsWithBoundary.length; i++ ) {
            jetAsyncItems.push( jetItemsWithBoundary[ i ].whenReady() );
        }
        
        // Setup min / max values / handlers. Note: Must be done after all item objects have been created
        // Note: Use apply to pass in jetAsyncItems as array of params
        $.when.apply( $, jetAsyncItems ).done( function() {
            initBoundaries( jetItemsWithBoundary );
        });
        
    } // attachDatePicker

    // Call to add main attach handler to the item interface
    item.addAttachHandler( attachDatePicker );

})( apex.item, apex.jQuery, apex.util, apex.lang );